# Don't touch! This is a keystring substituted in GitHub workflows!
__version__ = "1.0.0-rc.1"
###################################################################

from epx.job import Job, ModelConfig, ModelConfigSweep
from epx.run import Run, RunParameters
from epx.synthpop import SynthPop

# Set default logging handler to avoid "No handler found" warnings.
import logging
from logging import NullHandler

logging.getLogger(__name__).addHandler(NullHandler())



from datetime import datetime
import json
import logging
import re
from typing import (
    Dict,
    Iterable,
    List,
    Literal,
    NamedTuple,
    TypeAlias,
    TypeGuard,
    get_args,
)

import pandas as pd
import requests

from epx.run.exec.cloud.auth import platform_api_headers
from epx.run.exec.cloud.config import api_base_url
from epx.run.exec.cloud.strategy import ForbiddenResponse, UnauthorizedUserError
from epx.run.run import Run

logger = logging.getLogger(__name__)

StatusName = Literal["NOT STARTED", "RUNNING", "ERROR", "DONE"]
LogLevel = Literal["DEBUG", "INFO", "WARNING", "ERROR"]


class LogItem(NamedTuple):
    """An individual entry in the logs generated by FRED.

    Attributes
    ----------
    level : LogLevel
        The log level of the message, e.g. `INFO`, `ERROR`, etc.
    time : datetime
        The time that the message was reported at.
    message : str
        The log message.
    """

    level: LogLevel
    time: datetime
    message: str


RunWithId: TypeAlias = tuple[int, Run]


class JobStatus:
    def __init__(self, job_name: str, _run_with_ids: Iterable[RunWithId]):
        self.job_name = job_name
        self._run_with_ids = _run_with_ids
        self._log_re = re.compile(
            r"^\[(\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}.\d{3}Z)\] ([A-Z]*): (.*)$"
        )

    @property
    def logs(self) -> pd.DataFrame:
        """Return a collection of log entries output by FRED.

        Returns
        -------
        pd.DataFrame
            Collection of individual log entries generated by FRED during the
            run.
        """

        def process_log_line(line: str) -> LogItem:
            m = self._log_re.match(line)
            if m is None:
                raise ValueError(f"Invalid logline: {line}")
            level = m.group(2)
            assert self._is_valid_log_level(level)
            return LogItem(
                level,
                datetime.strptime(m.group(1), "%Y-%m-%dT%H:%M:%S.%fZ"),
                m.group(3),
            )

        data_frames: List[pd.DataFrame] = []
        for run_id, run in self._run_with_ids:
            try:
                p = run.output_dir / "logs.txt"
                with open(p, "r") as f:
                    log_records = tuple(
                        process_log_line(line) for line in f.readlines()
                    )
            except FileNotFoundError:
                log_records = tuple()
            logs = pd.DataFrame.from_records(
                log_records, columns=["level", "time", "message"]
            )
            data_frames.append(logs.assign(run_id=run_id))

        log_cols = ["run_id", "level", "time", "message"]
        if len(data_frames) != 0:
            return pd.concat(data_frames).loc[:, log_cols]
        return pd.DataFrame.from_records(tuple(), columns=log_cols)

    @property
    def name(self) -> StatusName:
        """Return a string summarizing the job status.

        Returns
        -------
        Status
            A string indicating the status of the job, one of: `"NOT STARTED"`,
            `"RUNNING"`, `"ERROR"`, or `"DONE"`.
        """

        pod_phase_map: Dict[str | None, StatusName] = {
            "Pending": "NOT STARTED",
            "Running": "RUNNING",
            "Succeeded": "DONE",
            "Failed": "ERROR",
            "Unknown": "ERROR",
            "": "NOT STARTED",
            None: "NOT STARTED",
        }

        # call api here to get StatusName
        endpoint_url = f"{api_base_url()}/runs"
        # Get request for a run to be executed to FRED Cloud API
        response = requests.get(
            endpoint_url,
            headers=platform_api_headers(),
            params={"job_name": str(self.job_name)},
        )

        # Check HTTP response status code and raise exceptions as appropriate
        if not response.ok:
            if response.status_code == requests.codes.forbidden:
                raise UnauthorizedUserError(
                    ForbiddenResponse.model_validate_json(response.text).description
                )
            else:
                raise RuntimeError(f"FRED Cloud error code: {response.status_code}")

        response_payload = json.loads(response.text)
        logger.debug(f"Payload: {response.text}")

        runs = response_payload["runs"]
        if len(runs) == 0:
            return "NOT STARTED"

        status_names = set(pod_phase_map[run["podPhase"]] for run in runs)

        if all(sn == "NOT STARTED" for sn in status_names):
            return "NOT STARTED"
        if "ERROR" in status_names:
            return "ERROR"
        if "RUNNING" in status_names:
            return "RUNNING"
        return "DONE"

    @staticmethod
    def _is_valid_log_level(level: str) -> TypeGuard[LogLevel]:
        """Helper method for validating that a string is a valid log level."""
        return level in get_args(LogLevel)

    def __repr__(self) -> str:
        return f"JobStatus({self.job_name})"

    def __str__(self) -> str:
        return self.name



"""Implementation for running FRED jobs using the Platform IDE.

The cloud execution strategy is implemented by the ``RunExecuteCloudStrategy``
class defined here. This class conforms to the ``RunExecuteStrategy``
interface.
"""

from datetime import date
import logging
from pathlib import Path
from typing import Optional, Union, List

import requests

from pydantic import BaseModel, ConfigDict, Field

from epx.synthpop import SynthPopModel, SynthPop
from epx.run.exec.common import (
    RunExecuteStrategy,
    RunParameters,
    RunExecuteMultipleStrategy,
    RunResponse,
    RunRequest,
    FREDArg,
)
from epx.run.exec.compat import fred_major_version
from epx.run.exec.cloud.auth import platform_api_headers
from epx.run.exec.cloud.config import api_base_url

logger = logging.getLogger(__name__)


class _RunRequestPayload(BaseModel):
    """The complete run request object passed to FRED Cloud.

    Attributes
    ----------
    run_requests : list[RunRequest]
        Collection of individual run request configurations.
    """

    model_config = ConfigDict(populate_by_name=True)

    run_requests: list[RunRequest] = Field(alias="runRequests")


class _RunResponseBody(BaseModel):
    """Response object for a batch of submitted runs from the /runs endpoint.

    Attributes
    ----------
    run_responses : list[RunResponse]
        Collection of responses for individual runs in the originating
        request.
    """

    model_config = ConfigDict(populate_by_name=True)

    run_responses: list[RunResponse] = Field(alias="runResponses")


class ForbiddenResponse(BaseModel):
    """Response object for a 403 Forbidden response from the SRS."""

    description: str


class RunExecuteCloudStrategy(RunExecuteStrategy):
    """Strategy for submitting an individual run to execute on FRED Cloud.

    Encapsulates logic for forming a request that is compatible with the FRED
    Cloud API /run endpoint based on user input, submitting that request
    to FRED Cloud, and converting any errors reported by FRED Cloud into Python
    exceptions.

    Parameters
    ----------
    job_name : str
        Name of the job to associate this run to on FRED Cloud.
    params : RunParameters
        Parameters to be passed to FRED configuring the run.
    output_dir : Path
        Directory in the user's environment to write run results to.
    size : str
        Name of instance size to use for the run.
    fred_version : str
        Version of FRED to use for the run.

    Attributes
    ----------
    job_name : str
        Name of the job to associate this run to on FRED Cloud.
    params : RunParameters
        Parameters to be passed to FRED configuring the run.
    output_dir : Path
        Directory in the user's environment to write run results to.
    size : str
        Name of instance size to use for the run.
    fred_version : str
        Version of FRED to use for the run.
    """

    def __init__(
        self,
        job_name: Union[str, None],
        params: RunParameters,
        output_dir: Path,
        size: str,
        fred_version: str,
    ):
        self.job_name = job_name
        self.params = params
        self.output_dir = output_dir
        self.size = size
        self.fred_version = fred_version

    def execute(self) -> RunResponse:
        """Submit a run to FRED Cloud.

        Raises
        ------
        UnauthorizedUserError
            If the user does not have sufficient privileges to perform the
            specified action on FRED Cloud.
        RuntimeError
            If a FRED Cloud server error occurs.
        IndexError
            If the FRED Cloud response unexpectedly implies the submission of
            multiple runs.
        RunConfigError
            If FRED Cloud reports an issue with the user's request.
        """
        endpoint_url = f"{api_base_url()}/runs"
        # Post request for a run to be executed to FRED Cloud API
        payload = self._request_payload().model_dump_json(by_alias=True)
        logger.debug(f"Request payload: {payload}")
        response = requests.post(
            endpoint_url,
            headers=platform_api_headers(),
            data=payload,
        )

        # Check HTTP response status code and raise exceptions as appropriate
        if not response.ok:
            if response.status_code == requests.codes.forbidden:
                raise UnauthorizedUserError(
                    ForbiddenResponse.model_validate_json(response.text).description
                )
            else:
                raise RuntimeError(f"FRED Cloud error code: {response.status_code}")
        response_payload = response.text
        logger.debug(f"Response payload: {response_payload}")
        response_body = _RunResponseBody.model_validate_json(response_payload)

        # Confirm response contains response data for exactly one run request
        if (response_length := len(response_body.run_responses)) != 1:
            raise IndexError(
                "Exactly 1 run request expected to be associated with response "
                f"but found {response_length}"
            )

        # Check for any run configuration errors reported in the response and
        # raise exceptions as appropriate
        errors = response_body.run_responses[0].errors
        if errors is not None:
            for error in errors:
                raise RunConfigError(error.key, error.error)

        return response_body.run_responses[0]

    def _request_payload(self) -> _RunRequestPayload:
        return _RunRequestPayload(
            run_requests=[
                RunRequest(
                    job_name=self.job_name,
                    working_dir=str(Path.cwd()),
                    size=self.size,
                    fred_version=self.fred_version,
                    population=(
                        SynthPopModel(
                            version=self.params.synth_pop.name,
                            locations=self.params.synth_pop.locations,
                        )
                        if self.params.synth_pop is not None
                        else None
                    ),
                    fred_args=(
                        _FREDArgsBuilder(self.fred_version)
                        .program(self.params.program)
                        .output_dir(self.output_dir.expanduser().resolve())
                        .overrides(self.params.model_params)
                        .seed(self.params.seed)
                        .start_date(self.params.start_date)
                        .end_date(self.params.end_date)
                        .locations(self.params.synth_pop)
                        .build()
                    ),
                )
            ]
        )


class RunExecuteMultipleCloudStrategy(RunExecuteMultipleStrategy):
    """Strategy for submitting multiple runs to execute on FRED Cloud.

    Encapsulates logic for forming a request that is compatible with the FRED
    Cloud API /run endpoint based on user input, submitting that request
    to FRED Cloud, and converting any errors reported by FRED Cloud into Python
    exceptions.
    """

    def __init__(
        self,
        runs,
    ):
        self.runs = runs

    def execute_all(self) -> List[RunResponse]:
        """Submit multiple runs to FRED Cloud.

        Raises
        ------
        UnauthorizedUserError
            If the user does not have sufficient privileges to perform the
            specified action on FRED Cloud.
        RuntimeError
            If a FRED Cloud server error occurs.
        IndexError
            If the FRED Cloud response unexpectedly implies the submission of
            multiple runs.
        RunConfigError
            If FRED Cloud reports an issue with the user's request.
        """
        endpoint_url = f"{api_base_url()}/runs"
        # Post request for a run to be executed to FRED Cloud API
        payload = self._request_payload_runs(self.runs).model_dump_json(by_alias=True)
        logger.debug(f"Request payload: {payload}")
        response = requests.post(
            endpoint_url,
            headers=platform_api_headers(),
            data=payload,
        )

        # Check HTTP response status code and raise exceptions as appropriate
        if not response.ok:
            if response.status_code == requests.codes.forbidden:
                raise UnauthorizedUserError(
                    ForbiddenResponse.model_validate_json(response.text).description
                )
            else:
                raise RuntimeError(f"FRED Cloud error code: {response.status_code}")
        response_payload = response.text
        logger.debug(f"Response payload: {response_payload}")
        response_body = _RunResponseBody.model_validate_json(response_payload)
        response_length = len(response_body.run_responses)

        # Confirm response contains response data for exactly one run request
        if response_length == 0:
            raise IndexError(
                "No responses received. Expected at least one response for run requests"
            )
        # Check for any run configuration errors reported in the response and
        # raise exceptions as appropriate
        for run_response in response_body.run_responses:
            errors = run_response.errors
            if errors is not None:
                for error in errors:
                    raise RunConfigError(error.key, error.error)
        return response_body.run_responses

    def _request_payload_runs(self, runs) -> _RunRequestPayload:
        return _RunRequestPayload(
            run_requests=[
                RunRequest(
                    job_name=run.job_name,
                    working_dir=str(Path.cwd()),
                    size=run.size,
                    fred_version=run.fred_version,
                    population=(
                        SynthPopModel(
                            version=run.params.synth_pop.name,
                            locations=run.params.synth_pop.locations,
                        )
                        if run.params.synth_pop is not None
                        else None
                    ),
                    fred_args=(
                        _FREDArgsBuilder(run.fred_version)
                        .program(run.params.program)
                        .output_dir(run.output_dir.expanduser().resolve())
                        .overrides(run.params.model_params)
                        .seed(run.params.seed)
                        .start_date(run.params.start_date)
                        .end_date(run.params.end_date)
                        .locations(run.params.synth_pop)
                        .build()
                    ),
                )
                for run in runs
            ]
        )


class _FREDArgsBuilder:
    """Builder for list of arguments to pass to FRED via SRS.

    Handles correct argument naming for different FRED versions.

    Parameters
    ----------
    fred_version : str
        FRED version for the run.
    """

    def __init__(self, fred_version: str):
        self.fred_version = fred_version
        self._args: list[FREDArg] = []

    def build(self) -> list[FREDArg]:
        return self._args

    def program(self, program: Union[Path, str]) -> "_FREDArgsBuilder":
        self._args.append(FREDArg(flag="-p", value=str(program)))
        return self

    def output_dir(self, output_dir: Path) -> "_FREDArgsBuilder":
        """Add the output directory argument to the FRED command line.

        Parameters
        ----------
        output_dir : Path
            Absolute path to the output directory.

        Returns
        -------
        _FREDArgsBuilder
            Current state of a partially built FRED command line.
        """
        if not output_dir.is_absolute():
            raise ValueError("output_dir must be an absolute path.")
        self._args.append(FREDArg(flag="-d", value=str(output_dir)))
        return self

    def overrides(
        self, model_params: Optional[dict[str, Union[float, str]]]
    ) -> "_FREDArgsBuilder":
        if model_params is not None:
            self._args.extend(
                [FREDArg(flag="-o", value=f"{k}={v}") for k, v in model_params.items()]
            )
        return self

    def seed(self, seed: int) -> "_FREDArgsBuilder":
        if fred_major_version(self.fred_version) < 11:
            # Use run number as an pseudo-seed
            self._args.append(FREDArg(flag="-r", value=str(seed)))
        else:
            self._args.append(FREDArg(flag="-s", value=str(seed)))
        return self

    def start_date(self, start_date: Optional[date]) -> "_FREDArgsBuilder":
        if start_date is not None:
            self._args.append(
                FREDArg(
                    flag="--start-date",
                    value=start_date.strftime(r"%Y-%m-%d"),
                )
            )
        return self

    def end_date(self, end_date: Optional[date]) -> "_FREDArgsBuilder":
        if end_date is not None:
            self._args.append(
                FREDArg(
                    flag="--end-date",
                    value=end_date.strftime(r"%Y-%m-%d"),
                )
            )
        return self

    def locations(self, synth_pop: Optional[SynthPop]) -> "_FREDArgsBuilder":
        if (
            synth_pop is not None
            and synth_pop.locations is not None
            and fred_major_version(self.fred_version) >= 11
        ):
            # FRED 10 does not support --locations flag
            for location in synth_pop.locations:
                self._args.append(FREDArg(flag="-l", value=location))
        return self


class UnauthorizedUserError(Exception):
    """Error indicating the user is not authorized to perform requested action.

    This can be thrown in the following circumstances:
        * User is not found in the Platform system
        * User does not have a subscription
        * User does not have sufficient compute credits

    Parameters
    ----------
    description : str
        Detailed description of the error as reported by SRS.

    Attributes
    ----------
    message : str
        Message describing the error.
    """

    def __init__(self, description: str):
        super().__init__(f"Authorization error: {description}")


class RunConfigError(Exception):
    """Error indicating there was a run configuration error.

    This is thrown when FRED Cloud API has reported an issue with the user's
    request.

    Parameters
    ----------
    key : str
        The general category of the error reported by FRED Cloud API, e.g.
        ``size`` for errors related to instance size, or ``fredVersion`` if the
        specified FRED version is not recognized.
    desc : str
        Detailed description of the error.

    Attributes
    ----------
    key : str
        The general category of the error reported by FRED Cloud API, e.g.
        ``size`` for errors related to instance size, or ``fredVersion`` if the
        specified FRED version is not recognized.
    desc : str
        Detailed description of the error.
    message : str
        Summary message describing the error.
    """

    def __init__(self, key: str, desc: str):
        self.key = key
        self.desc = desc
        super().__init__(f"{self.key} error: {self.desc}")



from typing import Optional

import numpy as np


def random_seed(meta_seed: Optional[int] = None) -> int:
    """Generate a RNG seed for the simulation engine.

    Used to generate a simulation seed for the user if they haven't specified
    one themselves.

    Parameters
    ----------
    meta_seed : Optional[int], optional
        Seed used to initialize the RNG that will itself be used to
        pseudo-randomly generate the simulation seed. If ``None`` (the default),
        unpredictable entropy from the OS is used in place of a ``meta_seed``.
        (see `docs`_ for ``numpy.random.default_rng``).

    Returns
    -------
    int
        Integer seed in range [0, 2**64).

    .. _docs: https://numpy.org/doc/stable/reference/random/generator.html#numpy.random.default_rng  # noqa

    Notes
    -----
    Generated seeds can take any value in the range that can be represented by a
    64-bit unsigned integer, matching the range of seed values that FRED can
    accept as input.
    """
    return int(np.random.default_rng(meta_seed).integers(0, 2**64, dtype=np.uint64))



from typing import Optional

from pydantic import BaseModel


class SynthPop:
    """An Epistemix synthetic population.

    Notes
    -----
    Currently this class is used to specify the synth pop locations for a given
    simulation run. Over time, methods will be added to support querying the
    synth pop independently of any particular run, analogously to the existing
    `epx-pop`_ package.

    .. _epx-pop: https://github.com/Epistemix-com/epx-pop/blob/main/epxpop/pop.py

    Parameters
    ----------
    name : str
        The name of a valid synthetic population, e.g. ``US_2010.v5``.
    locations : str
        List of location names within the specified synthetic population, e.g.
        ``["Allegheny_County_PA", "Jefferson_County_PA"]``.

    Attributes
    ----------
    name : str
        The name of a valid synthetic population, e.g. ``US_2010.v5``.
    locations : str
        List of location names within the specified synthetic population, e.g.
        ``["Allegheny_County_PA", "Jefferson_County_PA"]``.
    """

    def __init__(self, name: str, locations: Optional[list[str]]):
        self.name = name
        self.locations = locations

    def __repr__(self) -> str:
        return f"SynthPop(name={self.name}, locations={self.locations})"

    def __eq__(self, other) -> bool:
        if not isinstance(other, SynthPop):
            return False
        if (self.name == other.name) and (self.locations == other.locations):
            return True
        return False


class SynthPopModel(BaseModel):
    """Data interchange helper for synthetic population objects.

    This is in the format recognized by the /run endpoint, which differs
    slightly from the ``epx.synthpop.SynthPop`` class that is part of the
    user-facing API.

    Attributes
    ----------
    version : str
        Synthetic population version, e.g. ``US_2010.v5``
    locations : list[str]
        List of location names within the specified synthetic population, e.g.
        ``["Allegheny_County_PA", "Jefferson_County_PA"]``.
    """

    version: str
    locations: list[str]

    @staticmethod
    def from_synth_pop(synth_pop: SynthPop) -> "SynthPopModel":
        return SynthPopModel(
            version=synth_pop.name,
            locations=synth_pop.locations,
        )

    def as_synth_pop(self) -> SynthPop:
        return SynthPop(
            name=self.version,
            locations=self.locations,
        )



"""Functions for handling FRED 10 vs FRED 11 compatibility. """

import logging

import numpy as np

from epx.run.exec.common import RunParameters

logger = logging.getLogger(__name__)


def adapt_params_for_fred_version(
    params: RunParameters, fred_version: str
) -> RunParameters:
    """Adapt run parameters to be compatible with a specific FRED version.

    Parameters
    ----------
    params : RunParameters
        Run parameters to be adapted.
    fred_version : str
        FRED version to adapt the parameters for.

    Returns
    -------
    RunParameters
        Adapted run parameters.
    """
    if fred_major_version(fred_version) == 10:
        if params.seed is not None:
            seed = rescale_seed_to_run_number(params.seed)
            logger.info("Given seed rescaled to fit within FRED 10 run number range.")
        return RunParameters(
            program=params.program,
            synth_pop=params.synth_pop,
            start_date=params.start_date,
            end_date=params.end_date,
            model_params=params.model_params,
            seed=seed,
            compile_only=params.compile_only,
        )
    return params


def rescale_seed_to_run_number(seed: int) -> int:
    """Rescale seed to fit within the range of valid FRED run numbers.

    Parameters
    ----------
    seed : int
        Random seed on [0, 2**64) to be rescaled.

    Returns
    -------
    int
        Random run number on [1, max_run_number].

    Notes
    -----
    Depending on the platform it is compiled for, FRED stores the run number as
    either a 2 byte or 4 byte integer. Here we calculate the maximum allowed
    run number on the assumption that we can store a 2 byte integer.
    """
    max_run_number = 2 ** (8 * 2)
    rescaled_seed = np.interp(seed, [0, (2**64) - 1], [1, max_run_number])
    return int(rescaled_seed)


def fred_major_version(fred_version: str) -> int:
    """Extract the major version number from a FRED version string.

    Parameters
    ----------
    fred_version : str
        FRED version string.

    Returns
    -------
    int
        Major version number.
    """
    if fred_version == "latest":
        # TODO: Programmatically determine the latest version
        return 11
    return int(fred_version.split(".")[0])



from abc import ABC, abstractmethod
from pathlib import Path
import re
from typing import Iterable, Literal, NamedTuple, Optional, Union

CountType = Literal["count", "new", "cumulative"]


class OutputDirectoryError(Exception):
    def __init__(self, output_dir: Path, fred_version: Optional[str] = None):
        if fred_version is not None:
            message = (
                f"{output_dir} is not a valid FRED {fred_version} output directory."
            )
        else:
            message = f"{output_dir} is not a valid FRED output directory."
        super().__init__(message)


class VarBySimDayPath(NamedTuple):
    """Metadata about a file with the value of a variable on a sim day."""

    sim_day: int
    path: Path


class FileFinder(ABC):
    def __init__(self, run_output_dir: Path):
        """Base class for file finders used to find output across
        multiple versions of FRED.

        Abstracts filesystem locations to required files.

        Parameters
        ----------
        run_output_dir : Path
            The root directory for all files generated for a particular
            run. This directory should contain e.g. `conditions.json`,
            _NOT_ e.g. a directory called `RUN1`.
        """
        self.run_output_dir = run_output_dir

    @staticmethod
    def _validate_file_exists(p: Path) -> None:
        if not p.is_file():
            raise FileNotFoundError(f"File '{p}' not found.")

    def state(self, condition: str, state: str, count_type: CountType) -> Path:
        def get_prefix(count_type: CountType) -> str:
            match count_type:
                case "count":
                    return ""
                case "new":
                    return "new"
                case "cumulative":
                    return "tot"
                case _:
                    raise ValueError(f"State count type '{count_type}' invalid.")

        p = (
            self.run_output_dir
            / f"DAILY/{condition}.{get_prefix(count_type)}{state}.txt"
        )
        self._validate_file_exists(p)
        return p

    @property
    def dates(self) -> Path:
        p = self.run_output_dir / "DAILY/Date.txt"
        self._validate_file_exists(p)
        return p

    @property
    def epi_week(self) -> Path:
        p = self.run_output_dir / "DAILY/EpiWeek.txt"
        self._validate_file_exists(p)
        return p

    @property
    def pop_size(self) -> Path:
        p = self.run_output_dir / "DAILY/Popsize.txt"
        self._validate_file_exists(p)
        return p

    @property
    def conditions(self) -> Path:
        p = self.run_output_dir / "conditions.json"
        self._validate_file_exists(p)
        return p

    @property
    def return_code(self) -> Path:
        p = self.run_output_dir / "return_code.txt"
        self._validate_file_exists(p)
        return p

    @property
    @abstractmethod
    def print_output(self) -> Path: ...

    @abstractmethod
    def csv_output(self, filename: str) -> Path: ...

    @abstractmethod
    def text_output(self, filename: str) -> Path: ...

    @abstractmethod
    def numeric(self, varname: str) -> Path: ...

    @abstractmethod
    def list_table_by_simday(self, varname: str) -> Iterable[VarBySimDayPath]: ...

    @abstractmethod
    def table_by_simday(self, varname: str) -> Iterable[VarBySimDayPath]: ...

    @abstractmethod
    def network(self, network_name: str, sim_day: int) -> Path: ...


class FileFinderF10(FileFinder):
    """Accessor for filesystem paths of FRED output files for FRED <=10.1.0."""

    def __init__(self, output_dir: Path, run_number: Optional[int] = None):
        """Constructor for FileFinderF10.

        Parameters
        ----------
        output_dir : Path
            Path to simulation output directory corresponding to FRED's
            `--directory` argument.
        run_number : int, optional
            The run number to use for the simulation output directory. This
            argument is only required if the output directory does not yet
            exist, and is used to identify the expected run directory within
            the simulation output directory. By default `None`.

        Raises
        ------
        FileNotFoundError
            If a unique run directory could not be identified in the simulation
            output directory.
        """
        self.output_dir = output_dir
        super().__init__(self._get_run_output_dir(output_dir, run_number))

    @staticmethod
    def _get_run_output_dir(output_dir: Path, run_number: Optional[int] = None) -> Path:
        """Helper method to identify run directory within the simulation output
        directory.

        Notes
        -----
        In FRED <= 10.1.0, the simulation output directory contained
        (potentially) multiple runs. The name of each run directory was
        determined from FRED's ``--run-number`` argument, e.g. ``RUN1``,
        ``RUN12``, etc. This implementation requires that the simulation
        output directory contains only a single run.

        If the ``run_number`` argument is provided, the run directory is
        returned as ``<run_number>/RUN<run_number>``. Otherwise, we check
        for a single directory in the simulation output directory following
        the correct naming convention, and identify that as the run directory.

        Raises
        ------
        FileNotFoundError
            If a unique run directory could not be identified in the simulation
            output directory.
        """
        if run_number is not None:
            return output_dir / f"RUN{run_number}"
        candidate_run_dirs = [x for x in output_dir.glob("RUN*")]
        if len(candidate_run_dirs) != 1:
            raise FileNotFoundError(
                f"Could not identify unique run directory in {output_dir}"
            )
        return candidate_run_dirs[0]

    @property
    def print_output(self) -> Path:
        p = self.run_output_dir / "fred_out.txt"
        self._validate_file_exists(p)
        return p

    def csv_output(self, filename: str) -> Path:
        p = self.run_output_dir / f"CSV/{filename}"
        self._validate_file_exists(p)
        return p

    def text_output(self, filename: str) -> Path:
        p = self.run_output_dir / f"CSV/{filename}"
        self._validate_file_exists(p)
        return p

    def numeric(self, varname: str) -> Path:
        p = self.run_output_dir / f"DAILY/FRED.{varname}.txt"
        self._validate_file_exists(p)
        return p

    def list_table_end_of_sim(self, varname: str) -> Path:
        p = self.run_output_dir / f"LIST/{varname}.txt"
        self._validate_file_exists(p)
        return p

    def list_table_by_simday(self, varname: str) -> Iterable[VarBySimDayPath]:
        all_paths = self._var_by_simday_paths(varname)
        if len(all_paths) == 0:
            raise FileNotFoundError(
                f"No outputs for list table variable {varname} by sim day found."
            )
        return (x for x in all_paths)

    def table_end_of_sim(self, varname: str) -> Path:
        p = self.run_output_dir / f"LIST/{varname}.txt"
        self._validate_file_exists(p)
        return p

    def table_by_simday(self, varname: str) -> Iterable[VarBySimDayPath]:
        all_paths = self._var_by_simday_paths(varname)
        if len(all_paths) == 0:
            raise FileNotFoundError(
                f"No outputs for table variable {varname} by sim day found."
            )
        return (x for x in all_paths)

    def list_end_of_sim(self, varname: str) -> Path:
        p = self.run_output_dir / f"LIST/{varname}.txt"
        self._validate_file_exists(p)
        return p

    def list_by_simday(self, varname: str) -> Iterable[VarBySimDayPath]:
        all_paths = self._var_by_simday_paths(varname)
        if len(all_paths) == 0:
            raise FileNotFoundError(
                f"No outputs for list variable {varname} by sim day found."
            )
        return (x for x in all_paths)

    def network(self, network_name: str, sim_day: int) -> Path:
        p = self.run_output_dir / f"{network_name}-{sim_day}.vna"
        self._validate_file_exists(p)
        return p

    def _var_by_simday_paths(self, varname: str) -> list[VarBySimDayPath]:
        def get_sim_day(p: Path) -> int:
            m = re.match(r"\w*-(\d*).txt", p.name)
            if m is None:
                raise ValueError(f"{p} not recognized variable by sim day file")
            return int(m.group(1))

        all_paths = [
            VarBySimDayPath(get_sim_day(x), x)
            for x in (self.run_output_dir / "LIST").glob(f"{varname}-*.txt")
        ]
        all_paths.sort(key=lambda x: x.sim_day)
        return all_paths

    @property
    def errors(self) -> Optional[Path]:
        """Path to file containing details of any errors.

        Returns
        -------
        Path, optional
            Path to the errors file for the run, if one exists. Otherwise
            `None`.
        """
        p = self.run_output_dir / "errors.txt"
        try:
            self._validate_file_exists(p)
            return p
        except FileNotFoundError:
            return None

    @property
    def status(self) -> Path:
        p = self.run_output_dir / "status.txt"
        self._validate_file_exists(p)
        return p

    def __repr__(self) -> str:
        return f"FileFinderF10('{self.output_dir}')"


class FileFinderF11(FileFinder):
    """Accessor for filesystem paths of FRED output files for FRED >=11.0.0."""

    def __init__(self, output_dir: Path):
        """Constructor for FileFinderF11.

        Parameters
        ----------
        output_dir : Path
            Path to simulation output directory corresponding to FRED's
            `--directory` argument.
        """
        self.output_dir = output_dir
        super().__init__(output_dir)

    @property
    def print_output(self) -> Path:
        p = self.run_output_dir / "USER_OUTPUT/print_output.txt"
        self._validate_file_exists(p)
        return p

    def csv_output(self, filename: str) -> Path:
        p = self.run_output_dir / f"USER_OUTPUT/{filename}"
        self._validate_file_exists(p)
        return p

    def text_output(self, filename: str) -> Path:
        p = self.run_output_dir / f"USER_OUTPUT/{filename}"
        self._validate_file_exists(p)
        return p

    def numeric(self, varname: str) -> Path:
        p = self.run_output_dir / f"VARIABLES/numeric.{varname}.csv"
        self._validate_file_exists(p)
        return p

    def list_table_by_simday(self, varname: str) -> Iterable[VarBySimDayPath]:
        all_paths = self._var_by_simday_paths(varname, "list_table")
        if len(all_paths) == 0:
            raise FileNotFoundError(
                f"No outputs for list table variable {varname} by sim day found."
            )
        return (x for x in all_paths)

    def table_by_simday(self, varname: str) -> Iterable[VarBySimDayPath]:
        all_paths = self._var_by_simday_paths(varname, "table")
        if len(all_paths) == 0:
            raise FileNotFoundError(
                f"No outputs for table variable {varname} by sim day found."
            )
        return (x for x in all_paths)

    def list_(self, varname: str) -> Path:
        p = self.run_output_dir / f"VARIABLES/list.{varname}.csv"
        self._validate_file_exists(p)
        return p

    def network(self, network_name: str, sim_day: int) -> Path:
        p = self.run_output_dir / f"NETWORKS/{network_name}-{sim_day}.gv"
        self._validate_file_exists(p)
        return p

    def _var_by_simday_paths(self, varname: str, prefix: str) -> list[VarBySimDayPath]:
        def get_sim_day(p: Path) -> int:
            m = re.match(r"^[\w\.]*-(\d*)\.csv", p.name)
            if m is None:
                raise ValueError(f"{p} not recognized variable by sim day file")
            return int(m.group(1))

        all_paths = [
            VarBySimDayPath(get_sim_day(x), x)
            for x in (self.run_output_dir / "VARIABLES").glob(
                f"{prefix}.{varname}-*.csv"
            )
        ]
        all_paths.sort(key=lambda x: x.sim_day)
        return all_paths

    @property
    def logs(self) -> Path:
        """Path to log file.

        Returns
        -------
        Path
            Path to the log file for the run.
        """
        p = self.run_output_dir / "logs.txt"
        self._validate_file_exists(p)
        return p

    def __repr__(self) -> str:
        return f"FileFinderF11('{self.output_dir}')"


class FileFinderFactory:
    """Factory for selecting correct FileFinder implementation.

    Returns the appropriate FileFinder by inspecting the given output directory
    structure.

    Notes
    -----
        If the FRED `output_dir` provided to this class's constructor contains
        directories named like 'RUN*' then it is deemed to be a legacy
        results directory (FRED <= 10.1.0). Otherwise it is identified as a
        standard run results directory.

        run_number can be used to specify the expected location of a set of
        FRED<=10 results in case FRED has not already created the output
        directory at the time the FileFinder is instantiated.
    """

    def __init__(self, output_dir: Path, run_number: Optional[int | None] = None):
        self.output_dir = output_dir
        self.run_number = run_number

    def build(self) -> Union[FileFinderF10, FileFinderF11]:
        """Obtain a FileFinder appropriate for the given output directory.

        Notes
        -----
        We perform the following checks to determine the appropriate FileFinder:
        * If ``run_number`` was specified in the constructor, the output
          directory is assumed to be a legacy FRED output directory and a
          FileFinderF10 is returned.
        * If the output directory matches the structure of a FRED >= 11 results
          directory, a FileFinderF11 instance is returned.
        * If the output directory matches the structure of a FRED < 11 results
          directory, a FileFinderF10 instance is returned.
        * Otherwise, we assume the output directory is a yet-to-be-populated
          FRED >= 11 results directory.
        """
        if self.run_number is not None:
            return FileFinderF10(self.output_dir, self.run_number)
        elif self._is_output_dir(self.output_dir):
            return FileFinderF11(self.output_dir)
        elif self._is_legacy_output_dir(self.output_dir):
            return FileFinderF10(self.output_dir)
        else:
            return FileFinderF11(self.output_dir)

    @staticmethod
    def _is_legacy_output_dir(output_dir: Path) -> bool:
        candidate_run_dirs = [x for x in output_dir.glob("RUN*")]
        if len(candidate_run_dirs) != 0:
            return True
        return False

    @staticmethod
    def _is_output_dir(output_dir: Path) -> bool:
        try:
            if "conditions.json" in [x.name for x in output_dir.iterdir()]:
                return True
        except FileNotFoundError:
            return False
        return False



from abc import ABC, abstractmethod
from datetime import datetime
import itertools
from pathlib import Path
import re
from typing import Dict, Any, Iterable, Literal, NamedTuple, TypeGuard, Union, get_args

from pydantic import BaseModel, Field, ConfigDict
import pandas as pd

from epx.run.fs import FileFinderF11, FileFinderF10, FileFinder
from epx.run.exec.cloud.auth import platform_api_headers
from epx.run.exec.cloud.config import api_base_url
from epx.run.exec.cloud.strategy import (
    UnauthorizedUserError,
    ForbiddenResponse,
)
from epx.run.exec.common import RunRequest

import requests
import json

import logging

logger = logging.getLogger(__name__)

LogLevel = Literal["DEBUG", "INFO", "WARNING", "ERROR"]
StatusName = Literal["NOT STARTED", "RUNNING", "ERROR", "DONE"]


class _GetRunResponse(BaseModel):
    """Response object from the /runs endpoint for an individual run.

    Attributes
    ----------
    run_id : int
        Unique ID for the run.
    status : Literal["Submitted", "Failed"]
        Textual description of the status of the run.
    errors : list[RunError], optional
        List of any errors in the run configuration identified by the API
    run_request : _RunRequestPayload
        A copy of the originating request object that the response relates to.
    """

    model_config = ConfigDict(populate_by_name=True)

    id: int = Field(alias="id")
    user_id: int = Field(alias="userId")
    created_ts: datetime = Field(alias="createdTs")
    request: RunRequest = Field(alias="request")
    pod_phase: Literal["Pending", "Running", "Succeeded", "Failed", "Unknown"] = Field(
        alias="podPhase"
    )


class _GetRunsResponseBody(BaseModel):
    """Response object for a batch of submitted runs from the /runs endpoint.

    Attributes
    ----------
    run_responses : list[RunResponse]
        Collection of responses for individual runs in the originating
        request.
    """

    model_config = ConfigDict(populate_by_name=True)

    runs: list[_GetRunResponse] = Field(alias="runs")


class LogItem(NamedTuple):
    """An individual entry in the logs generated by FRED.

    Attributes
    ----------
    level : LogLevel
        The log level of the message, e.g. `INFO`, `ERROR`, etc.
    time : datetime
        The time that the message was reported at.
    message : str
        The log message.
    """

    level: LogLevel
    time: datetime
    message: str


class RunStatus(ABC):
    """Base class for returning run status information."""

    def __init__(self, file_finder: FileFinder, run_id: int | None):
        self._base_file_finder = file_finder
        self.run_id = run_id

    @property
    @abstractmethod
    def logs(self) -> pd.DataFrame:
        """Return a collection of log entries output by FRED.

        Returns
        -------
        pd.DataFrame
            Collection of individual log entries generated by FRED during the
            run.
        """
        ...

    @property
    def name(self) -> StatusName:
        """Return a string summarizing the run status.

        Returns
        -------
        Status
            A string indicating the status of the job, one of: `"NOT STARTED"`,
            `"RUNNING"`, `"ERROR"`, or `"DONE"`.
        """

        pod_phase_map: Dict[str, StatusName] = {
            "Pending": "NOT STARTED",
            "Running": "RUNNING",
            "Succeeded": "DONE",
            "Failed": "ERROR",
            "Unknown": "ERROR",
            "": "NOT STARTED",
        }

        # call API endpoint for that run
        endpoint_url = f"{api_base_url()}/runs"
        # Post request for a run to be executed to FRED Cloud API
        response = requests.get(
            endpoint_url,
            headers=platform_api_headers(),
            params={"id": str(self.run_id)},
        )
        logger.debug(f"Response: {response}")

        # Check HTTP response status code and raise exceptions as appropriate
        if not response.ok:
            if response.status_code == requests.codes.forbidden:
                raise UnauthorizedUserError(
                    ForbiddenResponse.model_validate_json(response.text).description
                )
            else:
                raise RuntimeError(f"FRED Cloud error code: {response.status_code}")

        # parse out the status
        response_payload = json.loads(response.text)
        logger.debug(f"Payload: {response.text}")

        if response_payload["runs"][0]["podPhase"] is None:
            return "NOT STARTED"
        return pod_phase_map[response_payload["runs"][0]["podPhase"]]

    def __str__(self):
        return self.name

    @staticmethod
    def _is_valid_log_level(level: str) -> TypeGuard[LogLevel]:
        """Helper method for validating that a string is a valid log level."""
        return level in get_args(LogLevel)


class RunStatusF10(RunStatus):
    """Accessor for information about the status of a run.

    Notes
    -----
    This implementation is for runs of FRED <= 10.1.0.
    """

    def __init__(self, file_finder: FileFinderF10, run_id: int | None):
        """LegacyRunStatus constructor.

        Parameters
        ----------
        file_finder : LegacyFileFinder
            Accessor for filesystem paths of FRED output files for FRED
            <=10.1.0.
        """
        super().__init__(file_finder, run_id)
        self._file_finder = file_finder
        self.run_id = run_id

    @property
    def logs(self) -> pd.DataFrame:
        try:
            log_iterators = [self._parse_status_file(self._file_finder.status)]
        except FileNotFoundError:
            log_iterators = []
        if self._file_finder.errors is not None:
            log_iterators.append(self._parse_errors_file(self._file_finder.errors))
        return pd.DataFrame.from_records(
            tuple(itertools.chain(*log_iterators)), columns=["level", "time", "message"]
        )

    @staticmethod
    def _parse_status_file(p: Path) -> Iterable[LogItem]:
        """Helper method to parse status.txt"""
        file_write_time = datetime.fromtimestamp(p.stat().st_mtime)
        with open(p, "r") as f:
            for line in f.readlines():
                trimmed_line = line.strip()
                if trimmed_line:
                    yield LogItem("INFO", file_write_time, trimmed_line)

    @staticmethod
    def _parse_errors_file(p: Path) -> Iterable[LogItem]:
        """Helper method to parse errors.txt."""
        file_write_time = datetime.fromtimestamp(p.stat().st_mtime)
        with open(p, "r") as f:
            messages = [x.strip() for x in f.read().split("FRED ERROR: ") if x]
        return tuple(LogItem("ERROR", file_write_time, m) for m in messages)

    def __repr__(self):
        return f"RunStatusF10({self._file_finder})"


class RunStatusF11(RunStatus):
    """Accessor for information about the status of a run.

    Notes
    -----
    This implementation is for runs of FRED >= 11.0.0.
    """

    def __init__(self, file_finder: FileFinderF11, run_id: int | None):
        super().__init__(file_finder, run_id)
        self._file_finder = file_finder
        self.run_id = run_id
        self._log_re = re.compile(
            r"^\[(\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}.\d{3}Z)\] ([A-Z]*): (.*)$"
        )

    @property
    def logs(self) -> pd.DataFrame:
        def process_log_line(line: str) -> LogItem:
            m = self._log_re.match(line)
            if m is None:
                raise ValueError(f"Invalid logline: {line}")
            level = m.group(2)
            assert self._is_valid_log_level(level)
            return LogItem(
                level,
                datetime.strptime(m.group(1), "%Y-%m-%dT%H:%M:%S.%fZ"),
                m.group(3),
            )

        try:
            with open(self._file_finder.logs, "r") as f:
                log_records = tuple(process_log_line(line) for line in f.readlines())
        except FileNotFoundError:
            log_records = tuple()
        return pd.DataFrame.from_records(
            log_records, columns=["level", "time", "message"]
        )

    def __repr__(self):
        return f"RunStatusF11({self._file_finder})"


class RunStatusFactory:
    """Factory for selecting correct RunStatus implementation."""

    def __init__(
        self, file_finder: Union[FileFinderF10, FileFinderF11], run_id: int | None
    ):
        self.file_finder = file_finder
        self.run_id = run_id

    def build(self) -> RunStatus:
        """Obtain a RunStatus appropriate for the given output directory.

        Raises
        ------
        TypeError
            If given `file_finder` is not a recognized type.
        """
        if self._is_legacy_file_finder(self.file_finder):
            return RunStatusF10(self.file_finder, self.run_id)
        elif self._is_file_finder(self.file_finder):
            return RunStatusF11(self.file_finder, self.run_id)
        else:
            raise TypeError(f"{self.file_finder} is not a valid FileFinder type")

    @staticmethod
    def _is_legacy_file_finder(file_finder: Any) -> TypeGuard[FileFinderF10]:
        if isinstance(file_finder, FileFinderF10):
            return True
        return False

    @staticmethod
    def _is_file_finder(file_finder: Any) -> TypeGuard[FileFinderF11]:
        if isinstance(file_finder, FileFinderF11):
            return True
        return False



import os
import logging

PLATFORM_API_VERSION = "0.4.0"


def api_base_url() -> str:
    try:
        return os.environ["PLATFORM_API_URL"]
    except KeyError:
        logging.debug("PLATFORM_API_URL environment variable not set.")
        return "https://studio.epistemix.cloud/v1"



"""Functionality for configuring and running collections runs (called jobs)."""

import io
import os

from dataclasses import dataclass
from datetime import date
from itertools import chain, islice, product, takewhile, repeat
import logging
from pathlib import Path
import shutil
from typing import Any, Iterable, List, Literal, Optional, TypeAlias, TypeVar, Union
import zipfile

import pandas as pd
from pandas._libs.tslibs.nattype import NaTType

from pydantic import BaseModel
import requests

from epx.config import get_cache_dir, default_results_dir
from epx.random import random_seed
from epx.run.exec.cloud.auth import platform_api_headers
from epx.run.exec.cloud.config import api_base_url

from epx.synthpop import SynthPop, SynthPopModel
from epx.run.run import Run, RunParameters
from epx.job.results import JobResults
from epx.job.status import JobStatus

from epx.run.exec.cloud.strategy import (
    ForbiddenResponse,
    RunExecuteMultipleCloudStrategy,
    UnauthorizedUserError,
)

import time

logger = logging.getLogger(__name__)

ModelParams: TypeAlias = dict[str, Union[float, str]]
PackedModelParams: TypeAlias = dict[
    str, Union[Iterable[Union[float, str]], Union[float, str]]
]
T = TypeVar("T")


class _SignedDownloadUrlInfo(BaseModel):
    """Response object of signed url.

    Attributes
     ----------
    run_id : str
        ID for the run.
    url: str
        The signed url from s3 for downloading job outputs.
    """

    run_id: int
    url: str


class _GetSignedDownloadUrlResponse(BaseModel):
    """Response collection of signed urls from the /job?job_name= endpoint."""

    urls: list[_SignedDownloadUrlInfo]


class DeleteOutcome(BaseModel):
    runId: int
    reason: Literal["Success", "NotFound", "Forbidden", "InternalError"]


class StopResponse(BaseModel):
    """Response object from the /runs endpoint for deleted SRS runs .

    Attributes
    ----------
    description : str
        The description of the status of the stop
    deletedIds: list[DeleteOutcome], optional
        List of runIds deleted successfully
    failedIds: list[DeleteOutcome], optional
        List of runIds deleted unsuccessfully
    """

    description: str
    deletedIds: Optional[list[DeleteOutcome]] = None
    failedIds: Optional[list[DeleteOutcome]] = None


@dataclass
class ModelConfig:
    """Configuration for a model run, including multiple realizations if
    applicable.

    Attributes
    ----------
    synth_pop : SynthPop
        Synthetic population to use for the run.
    start_date : Union[date, str], optional
        Simulation start date. If a ``str`` is given, should be in ISO 8601
        format, i.e. ``YYYY-MM-DD``.
    end_date : Union[date, str], optional
        Simulation end date. If a ``str`` is given, should be in ISO 8601
        format, i.e. ``YYYY-MM-DD``.
    model_params : ModelParams, optional
        Dictionary where the keys are model variable names and the values are
        the corresponding numeric or string values.
    seed : Union[int, Iterable[int]], optional
        Random number seeds for the configured runs. If ``None`` (the default),
        random seeds will be generated as required. If ``n_reps>1`` and a
        non-null value is given, this must be an iterable of length ``n_reps``.
    n_reps : int, optional
        Number of realizations of the model to run. By default, 1.
    """

    synth_pop: Optional[SynthPop] = None
    start_date: Optional[Union[date, str]] = None
    end_date: Optional[Union[date, str]] = None
    model_params: Optional[ModelParams] = None
    seed: Optional[Union[int, Iterable[int]]] = None
    n_reps: int = 1

    def __post_init__(self):
        self.seed = self._normalize_seed(self.seed, self.n_reps)

    @staticmethod
    def _normalize_seed(
        seed: Optional[Union[int, Iterable[int]]], n_reps: int
    ) -> Optional[Union[int, Iterable[int]]]:
        """Normalize and validate given seed value(s).

        Ensures that ``seed`` is compatible with the specified number of
        repetitions, ``n_reps``.
        """
        norm_seed: Optional[Union[int, Iterable[int]]]
        if seed is None:
            norm_seed = None
        else:
            if isinstance(seed, int):
                if n_reps == 1:
                    norm_seed = seed
                else:
                    raise IndexError(f"n_reps={n_reps} but a single seed given")
            elif isinstance(seed, Iterable):
                tuple_seeds = tuple(seed)
                if n_reps != len(tuple_seeds):
                    raise IndexError(
                        f"n_reps={n_reps} but {len(tuple_seeds)} seeds given"
                    )
                if len(tuple_seeds) == 1:
                    norm_seed = tuple_seeds[0]
                else:
                    norm_seed = tuple_seeds
        return norm_seed


class _ModelConfigModel(BaseModel):
    synth_pop: Optional[SynthPopModel] = None
    start_date: Optional[Union[date, str]] = None
    end_date: Optional[Union[date, str]] = None
    params: Optional[ModelParams] = None
    seed: Optional[Union[int, tuple[int, ...]]] = None
    n_reps: int = 1

    @staticmethod
    def from_model_config(model_config: "ModelConfig") -> "_ModelConfigModel":
        return _ModelConfigModel(
            synth_pop=(
                SynthPopModel.from_synth_pop(model_config.synth_pop)
                if model_config.synth_pop
                else None
            ),
            start_date=model_config.start_date,
            end_date=model_config.end_date,
            params=model_config.model_params,
            seed=model_config.seed,
            n_reps=model_config.n_reps,
        )

    def as_model_config(self) -> "ModelConfig":
        return ModelConfig(
            synth_pop=self.synth_pop.as_synth_pop() if self.synth_pop else None,
            start_date=self.start_date,
            end_date=self.end_date,
            model_params=self.params,
            seed=self.seed,
            n_reps=self.n_reps,
        )


class ModelConfigSweep:
    """Sweep over parameters for a given model.

    The set of model configurations represented by ``ModelConfigSweep`` objects
    constructed by taking the cartesian product of the given sequences of input
    parameters (see Examples).

    Parameters
    ----------
    synth_pop : Iterable[SynthPop]
        Sequence of populations to use for the runs.
    start_date : Iterable[Union[date, str]], optional
        Sequence of simulation start dates to use for the runs. If start dates
        are given as ``str``, they should be in ISO 8601 format, i.e.
        ``YYYY-MM-DD``.
    end_date : Iterable[Union[date, str]], optional
        Sequence of simulation end dates to use for the runs. If end dates
        are given as ``str``, they should be in ISO 8601 format, i.e.
        ``YYYY-MM-DD``.
    model_params : Iterable[PackedModelParams], optional
        Packed model parameters to use for the runs. Each item should be
        a dictionary where the keys are model variable names and the values
        can be a list of numeric or string values (float | str)
        or simply single numeric or string values. By default ``None``.
    seed : Union[Iterable[int], int], optional
        If an iterable is given, the number of elements must equal the number of
        combinations of values in the ``synth_pop``, ``start_date``,
        ``end_date``, and ``model_params`` iterables **multiplied** by
        ``n_reps``. If a single value is given, this is used as the 'meta seed'
        to pseudo-randomly generate seeds for each of the runs. If ``None`` is
        given, seeds for each run are generated using unpredictable entropy from
        the OS (see `docs`_ for ``numpy.random.default_rng``).
    n_reps : int, optional
        Number of realizations of each model configuration. By default, 1.

    .. _docs: https://numpy.org/doc/stable/reference/random/generator.html#numpy.random.default_rng  # noqa: E501

    Examples
    --------
    >>> sweep = ModelConfigSweep(
    ...     synth_pop=[SynthPop("US_2010.v5", ["Location1", "Location2"])],
    ...     start_date=["2024-01-01"],
    ...     end_date=["2024-01-31", "2024-02-29"],
    ...     model_params=[{"initialization_threshold": [0.95, 0.99], "ba_network": [0, 1]}]
    ... )
    >>> model_configs = list(sweep)
    >>> print(len(model_configs))
    8
    >>> print(model_configs[0].end_date)
    2024-01-31
    >>> print(model_configs[4].end_date)
    2024-02-29
    >>> print(model_configs[0].model_params)
    {'initialization_threshold': 0.95, 'ba_network': 0}
    """

    def __init__(
        self,
        synth_pop: Iterable[SynthPop],
        start_date: Optional[Iterable[Union[date, str]]],
        end_date: Optional[Iterable[Union[date, str]]],
        model_params: Optional[Iterable[PackedModelParams]] = None,
        seed: Optional[Union[Iterable[int], int]] = None,
        n_reps: int = 1,
    ):
        self.synth_pop = synth_pop
        self.start_date = self._normalize_optional_param(start_date)
        self.end_date = self._normalize_optional_param(end_date)
        self.model_params = self._unpack_model_params(
            self._normalize_optional_param(model_params)
        )
        self.seed = seed
        self.n_reps = n_reps
        self._configs = self._get_configs()

    @staticmethod
    def _normalize_optional_param(
        param: Optional[Iterable[T]],
    ) -> Iterable[Optional[T]]:
        """Ensure that ``param`` is an iterable of ``T``.

        Converting from Optional[Iterable[T]] to an Iterable[Optional[T]]
        allows us to pass the return value of this function to ``product``.
        Without this normalization, when ``None`` (rather than ``[None]``) was
        passed to ``product`` the returned iterable would be empty.
        """
        if param is None:
            return [None]
        return [x for x in param]

    def _unpack_model_params(
        self, model_params: Iterable[Optional[PackedModelParams]]
    ) -> Iterable[Optional[ModelParams]]:
        if not isinstance(model_params, list):
            raise ValueError("model_params should be a non-empty list of dictionaries")

        if model_params == [None]:
            return [None]

        unpacked_model_params = []

        for param in model_params:
            param_list = []
            for value in param.values():
                if isinstance(value, list):
                    param_list.append(value)
                else:
                    param_list.append([value])

            for combination in product(*param_list):
                unpacked_model_params.append(dict(zip(param.keys(), combination)))

        return unpacked_model_params

    def _get_configs(self) -> list[ModelConfig]:
        """Broadcast parameter combinations into a list of ModelConfigs."""
        configs = [
            x
            for x in product(
                self.synth_pop,
                self.start_date,
                self.end_date,
                self.model_params,
            )
        ]
        seeds: list[int]
        if self.seed is None:
            seeds = [random_seed() for _ in range(len(configs) * self.n_reps)]
        elif isinstance(self.seed, int):
            seeds = [random_seed(self.seed) for _ in range(len(configs) * self.n_reps)]
        else:
            try:
                iter(self.seed)
                seeds = [x for x in self.seed]
                if (n_seeds := len(seeds)) != (n_configs := len(configs) * self.n_reps):
                    raise IndexError(
                        f"Received {n_configs} configs but {n_seeds} seeds"
                    )
            except TypeError:
                raise ValueError("Invalid seed value")

        configs_with_seeds = zip(configs, self._split_every(self.n_reps, seeds))
        return [
            ModelConfig(
                synth_pop, start_date, end_date, model_params, seed, self.n_reps
            )
            for (
                synth_pop,
                start_date,
                end_date,
                model_params,
            ), seed in configs_with_seeds
        ]

    @staticmethod
    def _split_every(n: int, iterable: Iterable[Any]) -> Iterable[list[Any]]:
        """Slice an iterable into chunks of n elements.

        Parameters
        ----------
        n : int
            Number of elements in each chunk.
        iterable : Iterable
            Iterable to slice.

        Returns
        -------
        Iterator
            Iterator over the chunks of the input iterable.
        """
        iterator = iter(iterable)
        return takewhile(bool, (list(islice(iterator, n)) for _ in repeat(None)))

    def __iter__(self):
        return (x for x in self._configs)

    def __repr__(self) -> str:
        return (
            f"ModelConfigSweep("
            f"synth_pop={self.synth_pop}, "
            f"start_date={self.start_date}, "
            f"end_date={self.end_date}, "
            f"model_params={self.model_params}, "
            f"seed={self.seed}, "
            f"n_reps={self.n_reps}"
            f")"
        )


class _JobModel(BaseModel):
    program: Path
    config: list[_ModelConfigModel]
    key: str
    size: str = "hot"
    fred_version: str = "latest"
    n: int = 1
    results_dir: Optional[Path] = None

    @staticmethod
    def from_job(job: "Job") -> "_JobModel":
        return _JobModel(
            program=job.program,
            config=[_ModelConfigModel.from_model_config(x) for x in job.config],
            key=job.key,
            size=job.size,
            fred_version=job.fred_version,
            results_dir=job.results_dir,
        )

    def as_job(self) -> "Job":
        return Job(
            program=self.program,
            config=[x.as_model_config() for x in self.config],
            key=self.key,
            size=self.size,
            fred_version=self.fred_version,
            results_dir=self.results_dir,
        )


class Job:
    def __init__(
        self,
        program: Union[Path, str],
        config: Iterable[ModelConfig],
        key: str,
        size: str = "hot",
        fred_version="latest",
        results_dir: Optional[Union[Path, str]] = None,
        runIds=[],
        jobId: Optional[int] = None,
    ):
        """Client interface for configuring and running collections of
        simulation runs.

        Parameters
        ----------
        program : Union[Path, str]
            Path to the FRED entrypoint file.
        config : Iterable[ModelConfig]
            Set of model run configurations to execute.
        key : str
            Unique identifier for the job.
        size : str, optional
            Instance size to use for each run in the job, by default "hot".
        fred_version : str, optional
            FRED Simulation Engine version to use for each run in the job,
            by default "latest".
        results_dir : Optional[Union[Path, str]], optional
            Root results directory to use to store simulation results. By
            default ``None``, causing results to be stored in the default
            directory, ``~/results``.
        """
        self.program = Path(program)
        self.config = list(config)
        self.key = key
        self.size = size
        self.fred_version = fred_version
        self.results_dir = (
            Path(results_dir).expanduser().resolve()
            if results_dir
            else default_results_dir()
        )
        self.runIds = runIds
        self.jobId = jobId
        self._runs = self._build_runs(
            self.program,
            self.config,
            self.results_dir,
            self.key,
            self.size,
            self.fred_version,
        )

    @classmethod
    def _build_runs(
        cls,
        program: Path,
        config: Iterable[ModelConfig],
        results_dir: Path,
        key: str,
        size: str,
        fred_version: str,
    ) -> tuple[Run, ...]:
        def disaggregate_model_config(model_config: ModelConfig) -> list[ModelConfig]:
            """Convert model config representing multiple realizations into a
            list of model configs each representing a single realization.
            """
            if model_config.n_reps == 1:
                return [model_config]
            if isinstance(model_config.seed, Iterable):
                seeds: list[Optional[int]] = list(model_config.seed)
            else:
                seeds = [None for _ in range(model_config.n_reps)]
            return [
                ModelConfig(
                    synth_pop=model_config.synth_pop,
                    start_date=model_config.start_date,
                    end_date=model_config.end_date,
                    model_params=model_config.model_params,
                    seed=seeds[i],
                    n_reps=1,
                )
                for i, _ in enumerate(range(model_config.n_reps))
            ]

        def validate_singular_seed(
            seed: Optional[Union[int, Iterable[int]]]
        ) -> Optional[int]:
            if seed is not None and not isinstance(seed, int):
                raise ValueError("Seed must be an integer if n_reps=1")
            return seed

        job_dir = cls._get_job_output_dir(results_dir, key)
        job_dir.mkdir(parents=True, exist_ok=True)
        return tuple(
            Run(
                params=RunParameters(
                    program=program,
                    synth_pop=model_config.synth_pop,
                    start_date=model_config.start_date,
                    end_date=model_config.end_date,
                    model_params=model_config.model_params,
                    seed=validate_singular_seed(model_config.seed),
                ),
                output_dir=job_dir / str(run_id),
                size=size,
                fred_version=fred_version,
                job_name=key,
            )
            for run_id, model_config in enumerate(
                chain(
                    *[
                        disaggregate_model_config(model_config)
                        for model_config in config
                    ]
                )
            )
        )

    @staticmethod
    def _get_job_output_dir(results_dir: Path, key: str) -> Path:
        return results_dir / key

    @classmethod
    def from_key(cls, job_key: str) -> "Job":
        """Retrieve a Job object from a job key.

        Useful if one knows the key for a job but e.g. hadn't
        assigned the return value of `run_job` to a variable.

        Parameters
        ----------
        job_key : str
            The key of the job to retrieve.

        Raises
        ------
        JobKeyDoesNotExist
            If no job associated with the given job key exists yet.
        """
        try:
            job_config_file = cls._cache_dir(job_key) / "job.json"
            with open(job_config_file, "r") as f:
                return _JobModel.model_validate_json(f.read()).as_job()
        except FileNotFoundError as e:
            logger.error(e)
            raise JobDoesNotExistError(job_key)
        except ValueError as e:
            logger.error(e)
            raise

    @classmethod
    def list_keys(cls) -> List[str]:
        try:
            cache_dir = get_cache_dir() / "jobs"
            return [
                name
                for name in os.listdir(cache_dir)
                if os.path.isdir(os.path.join(cache_dir, name))
            ]
        except ValueError as e:
            logger.error(e)
            raise

    def execute(self, time_out: Optional[int] = None) -> None:
        """Execute the runs comprising the job.

        Parameters
        ----------
        time_out : int, optional
            The timeout of the job execution (in seconds).

        Raises
        ------
        RuntimeError:
            If the execution time exceeds timeout or
            If the execution occurs error.
        """
        self._verify_output_dir_empty()
        if time_out and (not isinstance(time_out, int) or time_out < 0):
            raise ValueError("Invalid timeout value")
        self._init_cache()
        self._write_job_config()
        start_time = time.time()  # start time of the job execution

        # Create a combined execution strategy
        exec_strategy_all = RunExecuteMultipleCloudStrategy(self._runs)

        for run in self._runs:
            run._verify_job_name()
            run._verify_output_dir_empty()
            run._init_cache()
        results = exec_strategy_all.execute_all()

        for run, result in zip(self._runs, results):
            run.run_id = result.run_id
            if result.run_id:
                self.runIds.append(result.run_id)
            run._write_run_config()

        if results[0] is not None:
            self.jobId = results[0].job_id

        if time_out:
            # time to wait (in seconds) before checking status again
            idle_time = 3
            while str(self.status) != "DONE":
                if str(self.status) == "ERROR":
                    logs = self.status.logs
                    log_msg = "; ".join(
                        logs.loc[logs.level == "ERROR"].message.tolist()
                    )
                    raise RuntimeError(
                        f"Job '{self.key}' failed with the following error:\n {log_msg}"
                    )
                elif time.time() > start_time + (time_out):
                    msg = f"Job did not finish within {time_out / 60} minutes."
                    raise RuntimeError(msg)
                time.sleep(idle_time)

    def stop(self) -> str:
        """Stop the running job.

        Users can only stop a job with the job status is RUNNING.

        Raises
        ------
        UnauthorizedUserError
            If the user does not have sufficient privileges to perform the
            specified action on FRED Cloud.
        RuntimeError
            If a FRED Cloud server error occurs.
            If the job status is different "RUNNING"
        """
        if str(self.status) != "RUNNING":
            msg = f"Can not stop the job with status is {self.status}."
            raise RuntimeError(msg)

        param = ""
        for index, id in enumerate(self.runIds):
            if index != len(self.runIds) - 1:
                param += f"id={id}&"
            else:
                param += f"id={id}"

        endpoint_url = f"{api_base_url()}/runs?{param}"
        # Patch request to delete SRS runs
        logger.debug(f"Request params: {param}")
        response = requests.patch(endpoint_url, headers=platform_api_headers())
        # Check HTTP response status code and raise exceptions as appropriate
        if not response.ok:
            if response.status_code == requests.codes.forbidden:
                raise UnauthorizedUserError(
                    ForbiddenResponse.model_validate_json(response.text).description
                )
            else:
                raise RuntimeError(f"FRED Cloud error code: {response.status_code}")
        response_payload = response.text
        response_body = StopResponse.model_validate_json(response_payload)

        return response_body.description

    def _write_job_config(self) -> None:
        with open(self._cache_dir(self.key) / "job.json", "w") as f:
            f.write(_JobModel.from_job(self).model_dump_json(by_alias=True))

    @staticmethod
    def _cache_dir(key: str) -> Path:
        return get_cache_dir() / "jobs" / key

    def _init_cache(self) -> None:
        self._cache_dir(self.key).mkdir(exist_ok=True, parents=True)

    def delete(self, interactive=True) -> None:
        """Delete all results data for the job.

        Users should be careful to ensure that the ``results_dir`` specified in
        the constructor is indeed the targeted run directory. This is a
        destructive operation and should be used with care. E.g. if
        ``results_dir = Path('/')`` this would cause the deletion of all files
        on the system that the user has write permissions for.

        Parameters
        ----------
        interactive : bool, optional
            Whether or not the ``delete`` command should be run interactively.
            When ``True`` (the default), the user will be prompted to confirm
            the deletion of the job results data. When ``False``, no
            confirmation prompt will be given. The latter option is provided to
            support programmatic usage, e.g. to delete the data for all jobs in
            a collection of jobs.
        """

        def confirm(key: str) -> bool:
            answer = input(f"Delete job '{key}'? [y/N]")
            if answer.lower() in ["y", "yes"]:
                return True
            else:
                return False

        def proceed():
            """
            Delete all run data and metadata caches if any
            """

            output_dir = self._get_job_output_dir(self.results_dir, self.key)
            cache_dir = self._cache_dir(self.key)
            paths = [output_dir, cache_dir]
            for path in paths:
                try:
                    if os.path.exists(path):
                        shutil.rmtree(path)
                except OSError:
                    raise RuntimeError(
                        f"An error occurred while deleting job {self.key}"
                    )
            print(f"Job {self.key} deleted successfully.")

        if not interactive or confirm(self.key):
            proceed()

    def _verify_output_dir_empty(self) -> None:
        """Ensure that ``self.results_dir/self.key`` does not contain any
        regular files.

        If ``self.results_dir/self.key`` does contain regular files, this is
        interpreted as meaning that a job of the given name already exists.

        Raises
        ------
        JobExistsError
            If the specified output_dir already contains regular files.
        """
        output_dir = self._get_job_output_dir(self.results_dir, self.key)
        if output_dir.is_dir():
            # output_dir exists
            if any(output_dir.iterdir()):
                # output_dir contains files
                raise JobExistsError(self.results_dir, self.key)

    @property
    def run_meta(self) -> pd.DataFrame:
        """Return metadata about each run in the job.

        Returns
        -------
        pd.DataFrame
            A DataFrame with columns:
                * ``run_id``: The index of the run in the job.
                * ``program``: The path to the FRED entrypoint file.
                * ``synth_pop``: The name of the synthetic population used.
                * ``locations``: The locations in the synthetic population.
                * ``start_date``: The start date of the simulation.
                * ``end_date``: The end date of the simulation.
                * ``params``: The model parameters.
                * ``seed``: The random seed used for the run.
                * ``size``: The instance size used for the run.
        """

        def proc_date(date: Optional[date]) -> Union[pd.Timestamp, NaTType]:
            return pd.Timestamp(date) if date is not None else pd.NaT

        return pd.DataFrame(
            {
                "run_id": run_id,
                "program": str(run.params.program),
                "synth_pop": (
                    run.params.synth_pop.name if run.params.synth_pop else None
                ),
                "locations": (
                    run.params.synth_pop.locations if run.params.synth_pop else None
                ),
                "start_date": proc_date(run.params.start_date),
                "end_date": proc_date(run.params.end_date),
                "params": run.params.model_params,
                "seed": run.params.seed,
                "size": run.size,
            }
            for run_id, run in enumerate(self._runs)
        )

    @property
    def status(self) -> JobStatus:
        """Current status of the job."""
        return JobStatus(
            self.key, ((run_id, run) for run_id, run in enumerate(self._runs))
        )

    @property
    def results(self) -> JobResults:
        """Object providing access to simulation results.
        Raises
        ------
        UnauthorizedUserError
            If the user does not have sufficient privileges to perform the
            specified action on FRED Cloud.
        RuntimeError
            If a FRED Cloud server error occurs.
            If the results do not exist in S3.
            If the Job is not DONE.
        """

        # To check if the results exist in the user's local results cache,
        # when a user attempts to interact with results.
        error_message = "Error occurred while accessing to simulation results"
        if self.jobId is None:
            raise RuntimeError(error_message)
        isExist = self._check_results_cache_exist(
            str(self._results_cache_dir(self.jobId))
        )
        if isExist is False:
            # If the Job is not DONE, throw an error
            if str(self.status) != "DONE":
                raise RuntimeError(error_message)
            # Get request to FRED Cloud API to get signed urls for downloading results
            signedUrls = self._get_signed_download_url(self.jobId).urls

            # To check if the number of runs is equal to the number of signed urls
            if len(signedUrls) != len(self._runs):
                raise RuntimeError(error_message)

            # To download output files from signed url and extract them
            for run_id, url in signedUrls:
                response = requests.get(url[1])

                # Check HTTP response status code and raise exceptions as appropriate
                if not response.ok:
                    raise RuntimeError(error_message)
                # Get file content and extract all on the fly
                with zipfile.ZipFile(io.BytesIO(response.content)) as zip_ref:
                    path = self._results_cache_dir(self.jobId) / str(run_id[1])
                    zip_ref.extractall(str(path))

        completed_run_results_with_ids = (
            (run_id, run.results)
            for run_id, run in enumerate(self._runs)
            if run.status.name == "DONE" and run.results is not None
        )
        return JobResults(completed_run_results_with_ids)

    @staticmethod
    def _results_cache_dir(key: int) -> Path:
        return Path.home() / ".epx/results_cache" / str(key)

    @staticmethod
    def _check_results_cache_exist(path: str) -> bool:
        try:
            if not os.listdir(path):
                return False
            return True
        except Exception:
            return False

    @staticmethod
    def _get_signed_download_url(key: int) -> _GetSignedDownloadUrlResponse:
        """Request to FRED Cloud API to get signed url for downloading job results.

        Raises
        ------
        UnauthorizedUserError
            If the user does not have sufficient privileges to perform the
            specified action on FRED Cloud.
        RuntimeError
        If a FRED Cloud server error occurs.
        """

        endpoint_url = f"{api_base_url()}/jobs"

        response = requests.get(
            endpoint_url,
            headers=platform_api_headers(),
            params={"job_id": key},
        )

        # Check HTTP response status code and raise exceptions as appropriate
        if not response.ok:
            if response.status_code == requests.codes.forbidden:
                raise UnauthorizedUserError(
                    ForbiddenResponse.model_validate_json(response.text).description
                )
            else:
                raise RuntimeError(f"FRED Cloud error code: {response.status_code}")

        response_payload = response.text
        logger.debug(f"Payload: {response.text}")
        response_body = _GetSignedDownloadUrlResponse.model_validate_json(
            response_payload
        )
        return response_body

    def __repr__(self) -> str:
        return (
            f"Job("
            f"program={self.program}, "
            f"config={self.config}, "
            f"key={self.key}, "
            f"size={self.size}, "
            f"fred_version={self.fred_version}, "
            f"results_dir={self.results_dir}"
            f")"
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Job):
            return False
        return (
            self.program == other.program
            and self.config == other.config
            and self.key == other.key
            and self.size == other.size
            and self.fred_version == other.fred_version
            and self.results_dir == other.results_dir
        )


class JobExistsError(Exception):
    """Raised when a job with a requested key already exists in the results."""

    def __init__(self, results_dir: Path, key: str):
        self.results_dir = results_dir
        self.key = key
        super().__init__(
            f"A job with key '{key}' already exists in results directory "
            f"'{results_dir}'"
        )


class JobDoesNotExistError(Exception):
    """Raised when a job with a requested key does not exist in the cache."""

    def __init__(self, key: str):
        self.key = key
        super().__init__(f"No job with key '{key}' exists")



"""Classes for combining and exposing results from multiple runs in a job."""

from typing import Iterable, TypeAlias, Optional

import pandas as pd

from epx.run.results import RunResults
from epx.run.fs import CountType

RunResultsWithId: TypeAlias = tuple[int, RunResults]


class JobResults:
    def __init__(self, run_results_with_ids: Iterable[RunResultsWithId]):
        """Results for all runs in a job.

        Parameters
        ----------
        run_results_with_ids : Iterable[RunResultsWithId]
            Iterable of ``RunResults`` objects that have been associated with
            run ids. These ids are used to associate results with their
            originating runs.

        Notes
        -----
        Calling code is responsible for checking that all ``RunResults`` are
        available, and none are ``None``.
        """
        self._run_results_with_ids = run_results_with_ids

    def state(self, condition: str, state: str, count_type: CountType) -> pd.DataFrame:
        """Return a time series of agent occupancy of a given FRED state.

        Parameters
        ----------
        condition : str
            Name of the FRED model condition containing the target state.
        state : str
            Name of the target state.
        count_type : CountType
            Type of count to report. Options are:
                * `count`, the number of agents occupying the state at the end
                  of each simulated day.
                * `new`, the number of agents entering the state at any time
                  during each simulated day.
                * `cumulative`, the cumulative number of times any agent has
                  entered the state since the beginning of the simulation,
                  reported at the end of each simulated day.

        Returns
        -------
        pd.DataFrame
            DataFrame with columns:
                * ``run_id``
                * ``sim_day``
                * The name of the requested count type, i.e. one of ``count``,
                  ``new``, or ``cumulative``.
        """

        return self._combine_results_across_runs(
            (
                (
                    run_result_with_id[0],
                    run_result_with_id[1].state(condition, state, count_type),
                )
                for run_result_with_id in self._run_results_with_ids
            )
        )

    def pop_size(self) -> pd.DataFrame:
        """Return a time series of population size during the run.

        Returns
        -------
        pd.DataFrame
            DataFrame with columns:
                * ``run_id``
                * ``sim_day``
                * ``pop_size``.
        """
        return self._combine_results_across_runs(
            (
                (
                    run_result_with_id[0],
                    run_result_with_id[1].pop_size(),
                )
                for run_result_with_id in self._run_results_with_ids
            )
        )

    def epi_weeks(self) -> pd.DataFrame:
        """Return a mapping from simulation days to epidemiological weeks.

        An epidemiological week, commonly referred to as an epi week or a CDC
        week, is a standardized method of counting weeks to allow for the
        comparison of data year after year.

        By definition, the first epi week of the year ends on the first Saturday
        of January that falls at least four days into the month. Each epi week
        begins on a Sunday and ends on a Saturday.

        Returns
        -------
        pd.DataFrame
            DataFrame with columns:
                * ``run_id``
                * ``sim_day``
                * ``epi_week`` with string values in ``YYYY.MM`` format.
        """

        return self._combine_results_across_runs(
            (
                (
                    run_result_with_id[0],
                    run_result_with_id[1].epi_weeks(),
                )
                for run_result_with_id in self._run_results_with_ids
            )
        )

    def dates(self) -> pd.DataFrame:
        """Return a mapping from simulation days to calendar dates.

        Returns
        -------
        pd.DataFrame
            DataFrame with columns:
                * ``run_id``
                * ``sim_day``
                * ``sim_date``
        """
        return self._combine_results_across_runs(
            (
                (
                    run_result_with_id[0],
                    run_result_with_id[1].dates(),
                )
                for run_result_with_id in self._run_results_with_ids
            )
        )

    def print_output(self) -> pd.DataFrame:
        """Return a series of messages output by FRED's ``print()`` action.

        Returns
        -------
        pd.DataFrame
            DataFrame with columns:
                * ``run_id``
                * ``print_output``
        """
        return self._combine_results_across_runs(
            (
                (
                    run_result_with_id[0],
                    run_result_with_id[1].print_output().to_frame(),
                )
                for run_result_with_id in self._run_results_with_ids
            )
        )

    def csv_output(self, filename: str) -> pd.DataFrame:
        """Return data output by FRED's ``print_csv`` action.

        Parameters
        ----------
        filename : str
            Name of output file as specified in the calls to ``print_csv`` in
            the FRED model code.

        Returns
        -------
        pd.DataFrame
            DataFrame containing data written to ``filename``. Columns are
            ``run_id`` (identifying the specific run that the output
            corresponds to), plus all columns specified in the call to
            ``open_csv`` in the model code.
        """
        return self._combine_results_across_runs(
            (
                (
                    run_result_with_id[0],
                    run_result_with_id[1].csv_output(filename),
                )
                for run_result_with_id in self._run_results_with_ids
            )
        )

    def file_output(self, filename: str) -> pd.DataFrame:
        """Return data output by FRED's ``print_file`` action.

        Parameters
        ----------
        filename : str
            Name of output file as specified in the calls to ``print_file`` in
            the FRED model code.

        Returns
        -------
        pd.DataFrame
            DataFrame with columns:
                * ``run_id``, identifying the specific run that the output
                  corresponds to.
                * ``file_output``, a string for entry output by calls to
                  ``print_file`` within the model code.
        """
        return self._combine_results_across_runs(
            (
                (
                    run_result_with_id[0],
                    run_result_with_id[1].file_output(filename).to_frame(),
                )
                for run_result_with_id in self._run_results_with_ids
            )
        )

    def numeric_var(self, varname: str) -> pd.DataFrame:
        """Return a time series of the value of a shared numeric variable.

        Parameters
        ----------
        varname : str
            Name of the shared numeric variable.

        Returns
        -------
        pd.DataFrame
            DataFrame with columns:
                * ``run_id``
                * ``sim_day``
                * ``value``
        """
        return self._combine_results_across_runs(
            (
                (
                    run_result_with_id[0],
                    run_result_with_id[1].numeric_var(varname),
                )
                for run_result_with_id in self._run_results_with_ids
            )
        )

    def list_var(self, varname: str, wide=False) -> pd.DataFrame:
        """Return a time series of the value of a shared list variable.

        Parameters
        ----------
        varname : str
            Name of the shared list variable.
        wide : bool, optional
            Return data in 'wide' format where a column is created for each
            list element. Defaults to ``False``.

        Returns
        -------
        pd.DataFrame
            DataFrame with columns:
                * ``run_id``
                * ``sim_day``
                * ``list_index``
                * ``value``
            if ``wide=False``. Otherwise:
                * ``run_id``
                * ``sim_day``
                * ``item_0``
                * ``item_1``
                * ...
        """
        return self._combine_results_across_runs(
            (
                (
                    run_result_with_id[0],
                    run_result_with_id[1].list_var(varname, wide=wide),
                )
                for run_result_with_id in self._run_results_with_ids
            )
        )

    def list_table_var(self, varname: str, wide=False) -> pd.DataFrame:
        """Return a time series of the value of a shared list_table variable.

        Parameters
        ----------
        varname : str
            Name of the shared list_table variable.
        wide : bool, optional
            Return data in 'wide' format where a column is created for each
            list element. Defaults to ``False``.

        Returns
        -------
        pd.DataFrame
            DataFrame with columns:
                * ``run_id``
                * ``sim_day``
                * ``key``
                * ``list_index``
                * ``value``
            if ``wide=False``. Otherwise:
                * ``run_id``
                * ``sim_day``
                * ``key``
                * ``item_0``
                * ``item_1``
                * ...
        """
        return self._combine_results_across_runs(
            (
                (
                    run_result_with_id[0],
                    run_result_with_id[1].list_table_var(varname, wide=wide),
                )
                for run_result_with_id in self._run_results_with_ids
            )
        )

    def table_var(self, varname: str) -> pd.DataFrame:
        """Return a time series of the value of a shared table variable.

        Parameters
        ----------
        varname : str
            Name of the shared table variable.

        Returns
        -------
        pd.DataFrame
                * ``run_id``
                * ``sim_day``
                * ``key``
                * ``value``
        """
        return self._combine_results_across_runs(
            (
                (
                    run_result_with_id[0],
                    run_result_with_id[1].table_var(varname),
                )
                for run_result_with_id in self._run_results_with_ids
            )
        )

    def network(
        self, network_name: str, is_directed: bool = True, sim_day: Optional[int] = None
    ) -> pd.Series:
        """Return a series of network objects from a job.

        Parameters
        ----------
        network_name : str
            Name of the FRED network.
        is_directed : bool, optional
            Whether the network is directed. Defaults to ``True``.
        sim_day : Optional[int], optional
            Simulation day to return the network for. If ``None``, return the
            network for the final simulation day. Defaults to ``None``.

        Returns
        -------
        pd.Series
            Series with the network objects from the runs within the job.
        """
        sim_day_network_map = {
            run_result_with_id[0]: run_result_with_id[1].network(
                network_name, is_directed, sim_day
            )
            for run_result_with_id in self._run_results_with_ids
        }

        return pd.Series(sim_day_network_map, name=network_name)

    @staticmethod
    def _combine_results_across_runs(
        results_dfs_with_ids: Iterable[tuple[int, pd.DataFrame]]
    ) -> pd.DataFrame:
        def proc_result_df(run_id: int, df: pd.DataFrame) -> pd.DataFrame:
            cols = list(df.columns)
            cols.insert(0, "run_id")
            return df.assign(run_id=run_id).loc[:, cols]

        return pd.concat(
            [proc_result_df(run_id, df) for run_id, df in results_dfs_with_ids]
        ).reset_index(drop=True)



import os
import logging
from pathlib import Path
from typing import Optional, Union
from datetime import datetime

from pydantic import BaseModel

from epx.config import get_cache_dir
from epx.run.fs import FileFinderFactory
from epx.run.status import RunStatus, RunStatusFactory
from epx.run.results import RunResults, RunResultsFactory
from epx.run.exec.compat import (
    adapt_params_for_fred_version,
    fred_major_version,
    rescale_seed_to_run_number,
)
from epx.run.exec import (
    RunExecuteStrategy,
    RunExecuteCloudStrategy,
    RunParameters,
    RunParametersModel,
)

logger = logging.getLogger(__name__)


class _RunModel(BaseModel):
    job_name: Union[str, None]
    run_id: int | None
    params: RunParametersModel
    output_dir: Union[Path, str]
    size: str
    fred_version: str

    @staticmethod
    def from_run(run: "Run") -> "_RunModel":
        return _RunModel(
            job_name=run.job_name,
            run_id=run.run_id,
            params=RunParametersModel.from_run_parameters(run.params),
            output_dir=run.output_dir,
            size=run.size,
            fred_version=run.fred_version,
        )

    def as_run(self) -> "Run":
        return Run._from_file(
            job_name=self.job_name if hasattr(self, "job_name") else None,
            run_id=self.run_id,
            params=self.params.as_run_parameters(),
            output_dir=self.output_dir,
            size=self.size,
            fred_version=self.fred_version,
        )


class Run:
    """Client interface for manipulating individual simulation runs.

    Parameters
    ----------
    params : RunParameters
        Parameters to be passed to FRED configuring the run.
    output_dir : Union[Path, str]
        Directory in the user's environment to write run results to.
    size : str, optional
        Size of cloud instance to use for the run. Defaults to ``hot``.
    fred_version : str, optional
        FRED version to use for the run. Defaults to ``latest``.
    job_name : str, optional
        Name of the job. If not specified, defaults to the current datetime.

    Attributes
    ----------
    run_id : int, optional
        The runId returned from the API on execute.
    params : RunParameters
        Parameters to be passed to FRED configuring the run.
    output_dir : Path
        Directory in the user's environment to write run results to.
    size : str, optional
        Size of cloud instance to use for the run.
    fred_version : str, optional
        FRED version to use for the run.
    job_name : str, optional
        Name of the job. If not specified, defaults to the current datetime.
    """

    def __init__(
        self,
        params: RunParameters,
        output_dir: Union[Path, str],
        size: str = "hot",
        fred_version: str = "latest",
        job_name: Union[str, None] = None,
    ):
        self.job_name = job_name if job_name is not None else str(datetime.now())
        self.run_id: int | None = None
        self.params = params
        self.output_dir = Path(output_dir).expanduser().resolve()
        self.size = size
        self.fred_version = fred_version
        self._exec_strategy: RunExecuteStrategy = RunExecuteCloudStrategy(
            self.job_name,
            adapt_params_for_fred_version(self.params, fred_version),
            self.output_dir,
            self.size,
            self.fred_version,
        )

    @staticmethod
    def _from_file(
        params: RunParameters,
        output_dir: Union[Path, str],
        size: str = "hot",
        fred_version: str = "latest",
        job_name: Union[str, None] = None,
        run_id: int | None = None,
    ) -> "Run":
        run = Run(params, output_dir, size, fred_version, job_name)
        run.run_id = run_id
        return run

    @classmethod
    def from_output_dir(cls, output_dir: Union[Path, str]) -> "Run":
        """Construct a ``Run`` object from a previously executed simulation run.

        Uses data from the ``output_dir`` to obtain a Run object.

        Parameters
        ----------
        output_dir : Union[Path, str]
            The output directory used for the previously executed run.
        """
        try:
            run_config_file = cls._cache_dir(Path(output_dir)) / "run.json"
            with open(run_config_file, "r") as f:
                run_model = _RunModel.model_validate_json(f.read())
                if not hasattr(run_model, "job_name") or run_model.job_name is None:
                    run_model.job_name = None
                return run_model.as_run()
        except FileNotFoundError as e:
            logger.error(e)
            raise
        except ValueError as e:
            logger.error(e)
            raise

    @staticmethod
    def _cache_dir(output_dir: Path) -> Path:
        """Construct path to the directory in which to store cached data for the
        run.

        The absolute path to the output directory is the unique name for the
        run. Cached data for each run is stored under
        ``~/.epx_client/runs/<run_name>``.

        Parameters
        ----------
        output_dir : Path
            Directory in the user's environment to write run results to.

        Notes
        -----
        We considered using ``output_dir/epx_client`` as the location for the
        run cache. However, currently FRED Cloud silently fails if the specified
        output directory already contains data when the run request is made.
        Storing run cache data separately in ``~/.epx_client/runs/`` avoids this
        problem.
        """
        run_name = str(output_dir.resolve())[1:]
        return get_cache_dir() / "runs" / Path(run_name)

    def execute(self) -> None:
        """Execute the simulation run."""
        self._verify_job_name()
        self._verify_output_dir_empty()
        self._init_cache()
        result = self._exec_strategy.execute()
        self.run_id = result.run_id
        self._write_run_config()

    def _verify_job_name(self) -> None:
        """Ensure that ``job_name`` is specified.

        This is required to make this code backwards compatible with runs created
        before the introduction of the ``job_name`` parameter.

        Raises
        ------
        ValueError
            If ``job_name`` is not specified.
        """
        if self.job_name is None:
            raise ValueError("job_name must be specified")

    def _verify_output_dir_empty(self) -> None:
        """Ensure that ``output_dir`` does not contain any regular files.

        If ``output_dir`` does contain regular files, this is interpreted as
        meaning that a run of the given name already exists.

        Raises
        ------
        RunExistsError
            If the specified output_dir already contains regular files.
        """
        if self.output_dir.is_dir():
            # output_dir exists
            if any(self.output_dir.iterdir()):
                # output_dir contains files
                raise RunExistsError(self.output_dir)

    def _init_cache(self) -> None:
        self._cache_dir(self.output_dir).mkdir(exist_ok=True, parents=True)

    def _write_run_config(self) -> None:
        with open(self._cache_dir(self.output_dir) / "run.json", "w") as f:
            f.write(_RunModel.from_run(self).model_dump_json())

    def delete(self, interactive=True) -> None:
        """Delete all results data for the run.

        Users should be careful to ensure that the ``output_dir`` specified in
        the constructor is indeed the targeted run directory. This is a
        destructive operation and should be used with care. E.g. if
        ``output_dir = Path('/')`` this would cause the deletion of all files
        on the system that the user has write permissions for.

        Parameters
        ----------
        interactive : bool, optional
            Whether or not the ``delete`` command should be run interactively.
            When ``True`` (the default), the user will be prompted to confirm
            the deletion of all files in the run directory. When ``False``, no
            confirmation prompt will be given. The latter option is provided to
            support programmatic usage, e.g. to delete the data for all runs in
            a collection of runs.
        """

        def confirm(output_dir: Path) -> bool:
            answer = input(f"Delete contents of {output_dir}? [y/N]")
            if answer.lower() in ["y", "yes"]:
                return True
            else:
                return False

        def proceed():
            for root, dirs, files in os.walk(self.output_dir, topdown=False):
                for name in files:
                    (Path(root) / name).unlink()
                for name in dirs:
                    (Path(root) / name).rmdir()
            self.output_dir.rmdir()

        if not interactive or confirm(self.output_dir):
            proceed()

    @property
    def status(self) -> RunStatus:
        """Status object for the run."""
        return RunStatusFactory(
            FileFinderFactory(self.output_dir).build(), run_id=self.run_id
        ).build()

    @property
    def results(self) -> Optional[RunResults]:
        """RunResults, optional: Results object for the run. If the results
        directory has not been populated yet, returns ``None``.
        """
        if str(self.status) != "DONE":
            return None
        if fred_major_version(self.fred_version) < 11:
            run_number = rescale_seed_to_run_number(self.params.seed)
        else:
            run_number = None
        return RunResultsFactory(
            FileFinderFactory(self.output_dir, run_number).build()
        ).build()

    def __repr__(self) -> str:
        return (
            f"Run("
            f"job_name={self.job_name}, "
            f"params={self.params}, "
            f"output_dir={self.output_dir}, "
            f"size={self.size}, "
            f"fred_version={self.fred_version}"
            f")"
        )

    def __eq__(self, other) -> bool:
        if not isinstance(other, Run):
            return False
        if (
            (self.params == other.params)
            and (self.output_dir == other.output_dir)
            and (self.size == other.size)
            and (self.fred_version == other.fred_version)
        ):
            return True
        return False


class RunExistsError(Exception):
    """Thrown when user specifies an output_dir that already contains data."""

    def __init__(self, output_dir: Path):
        self.output_dir = output_dir
        super().__init__(
            f"Run data already exists in output_dir: {self.output_dir}. "
            "Call Run.delete to delete this data and reuse output_dir."
        )



"""Global configuration for the EPX client."""

import os
from pathlib import Path


def get_cache_dir() -> Path:
    """Return the path to the cache directory for the EPX client.

    This is used to store metadata about runs and jobs initiated through the
    client. If the environment variable ``EPX_CACHE_DIR`` is set, this is used.
    Otherwise, the default location is ``~/.epx_client``.
    """
    try:
        return Path(os.environ["EPX_CACHE_DIR"])
    except KeyError:
        return Path.home().resolve() / ".epx_client"


def default_results_dir() -> Path:
    return Path.home() / "results"



"""Structures used by all run execution strategies."""

from datetime import date, datetime
from pathlib import Path
from typing import Protocol, Optional, Union, Literal, List

from pydantic import BaseModel, Field, ConfigDict

from epx.random import random_seed
from epx.synthpop import SynthPop, SynthPopModel


class FREDArg(BaseModel):
    """A FRED command line argument/ value pair.

    Attributes
    ----------
    flag : str
        The command line flag to pass to FRED, e.g. ``-p``.
    value : str
        The value corresponding to the command line flag, e.g.
        ``my-model/main.fred``.
    """

    flag: str
    value: str


class RunRequest(BaseModel):
    """Configuration for an individual run to be executed.

    Attributes
    ----------
    working_dir : str
        Working directory that FRED should be called from. This ensures
        that e.g relative paths within the model code are resolved correctly.
    size : str
        Name of instance size to use for the run.
    fred_version : str
        Version of FRED to use for the run.
    population : SynthPopModel
        The specific locations within a synthetic population that should be
        used for the simulation.
    fred_args : list[FREDArg]
        Command line arguments to be passed to FRED.
    """

    model_config = ConfigDict(populate_by_name=True)

    job_name: str = Field(alias="jobName")
    working_dir: str = Field(alias="workingDir")
    size: str
    fred_version: str = Field(alias="fredVersion")
    population: Optional[SynthPopModel] = None
    fred_args: list[FREDArg] = Field(alias="fredArgs")


class RunError(BaseModel):
    """A run configuration error in FRED Cloud responses.

    Attributes
    ----------
    key : str
        The general category of the error reported by FRED Cloud API, e.g.
        ``size`` for errors related to instance size, or ``fredVersion`` if the
        specified FRED version is not recognized.
    error : str
        Detailed description of the error.
    """

    key: str
    error: str


class RunResponse(BaseModel):
    """Response object from the /runs endpoint for an individual run.

    Attributes
    ----------
    run_id : int
        Unique ID for the run.
    status : Literal["Submitted", "Failed"]
        Textual description of the status of the run.
    errors : list[RunError], optional
        List of any errors in the run configuration identified by the API
    run_request : _RunRequestPayload
        A copy of the originating request object that the response relates to.
    """

    model_config = ConfigDict(populate_by_name=True)

    run_id: int = Field(alias="runId")
    job_id: int = Field(alias="jobId")
    status: Literal["Submitted", "Failed"]
    errors: Optional[list[RunError]] = None
    run_request: RunRequest = Field(alias="runRequest")


class RunParameters:
    """Parameters to configure a run.

    Notes
    -----
    In a future version of the client, we plan to support ``program`` being
    specified with type ``Union[Path, list[Path], str, list[str]]``. This will
    make it possible for users to avoid specifying a single entrypoint file
    and instead provide an ordered list of ``.fred`` model files to include.

    Parameters
    ----------
    program : Union[Path, str]
        FRED entrypoint filename.
    synth_pop : SynthPop
        Synthetic population to use for the run.
    start_date : Union[date, str], optional
        Simulation start date. If a ``str`` is given, should be in ISO 8601
        format, i.e. ``YYYY-MM-DD``.
    end_date : Union[date, str], optional
        Simulation end date. If a ``str`` is given, should be in ISO 8601
        format, i.e. ``YYYY-MM-DD``.
    model_params : dict[str, Union[float, str]], optional
        Dictionary where the keys are model variable names and the values are
        the corresponding numeric or string values.
    seed : int, optional
        Random number seed for the run. If ``None``, a random seed will be
        generated.
    compile_only : bool, optional
        If ``True``, compile the FRED model, but do not run it. Defaults to
        ``False``.

    Attributes
    ----------
    program : Path
        FRED entrypoint filename.
    synth_pop : SynthPop
        Synthetic population to use for the run.
    start_date : date
        Simulation start date.
    end_date : date
        Simulation end date.
    model_params : dict[str, Union[float, str]], optional
        Dictionary where the keys are model variable names and the values are
        the corresponding numeric or string values.
    seed : int
        Random number seed for the run.
    compile_only : bool
        If ``True``, compile the FRED model, but do not run it. Defaults to
        ``False``.
    """

    def __init__(
        self,
        program: Union[Path, str],
        synth_pop: Optional[SynthPop] = None,
        start_date: Optional[Union[date, str]] = None,
        end_date: Optional[Union[date, str]] = None,
        model_params: Optional[dict[str, Union[float, str]]] = None,
        seed: Optional[int] = None,
        compile_only: bool = False,
    ):
        self.program = Path(program)
        self.synth_pop = synth_pop
        self.start_date = self._normalize_date(start_date) if start_date else None
        self.end_date = self._normalize_date(end_date) if end_date else None
        self.model_params = model_params
        self.seed: int = seed if seed is not None else random_seed()
        self.compile_only = compile_only

    @staticmethod
    def _normalize_date(d: Union[date, str]) -> date:
        if isinstance(d, date):
            return d
        elif isinstance(d, str):
            return datetime.strptime(d, r"%Y-%m-%d").date()
        else:
            raise TypeError(f"Date format not recognized: {d}")

    def __repr__(self) -> str:
        return (
            f"RunParameters("
            f"program={self.program}, "
            f"synth_pop={self.synth_pop}, "
            f"start_date={self.start_date}, "
            f"end_date={self.end_date}, "
            f"model_params={self.model_params}, "
            f"seed={self.seed}, "
            f"compile_only={self.compile_only}"
            f")"
        )

    def __eq__(self, other) -> bool:
        if not isinstance(other, RunParameters):
            return False
        if (
            (self.program == other.program)
            and (self.synth_pop == other.synth_pop)
            and (self.start_date == other.start_date)
            and (self.end_date == other.end_date)
            and (self.model_params == other.model_params)
            and (self.seed == other.seed)
            and (self.compile_only == other.compile_only)
        ):
            return True
        return False


class RunParametersModel(BaseModel):
    """Data model facilitating JSON serialization of ``RunParameters`` objects.

    The reason for having both a Pydantic model (this class) and a vanilla
    Python class (``RunParameters``) is that Pydantic models' constructors
    require arguments to be passed as keywords, and their method signatures
    (i.e. those exposed by the builtin ``help`` function) do not explicitly show
    required arguments and types. ``RunParameters`` is part of the public API,
    and these limitations would make Pydantic models awkward for users to
    interact with.
    """

    program: Union[Path, str]
    synth_pop: Optional[SynthPopModel] = None
    start_date: Optional[Union[date, str]]
    end_date: Optional[Union[date, str]]
    sim_model_params: Optional[dict[str, Union[float, str]]] = None
    seed: Optional[int] = None
    compile_only: bool = False

    @staticmethod
    def from_run_parameters(run_parameters: RunParameters) -> "RunParametersModel":
        return RunParametersModel(
            program=run_parameters.program,
            synth_pop=(
                SynthPopModel.from_synth_pop(run_parameters.synth_pop)
                if run_parameters.synth_pop
                else None
            ),
            start_date=run_parameters.start_date,
            end_date=run_parameters.end_date,
            sim_model_params=run_parameters.model_params,
            seed=run_parameters.seed,
            compile_only=run_parameters.compile_only,
        )

    def as_run_parameters(self) -> RunParameters:
        return RunParameters(
            program=self.program,
            synth_pop=self.synth_pop.as_synth_pop() if self.synth_pop else None,
            start_date=self.start_date,
            end_date=self.end_date,
            model_params=self.sim_model_params,
            seed=self.seed,
            compile_only=self.compile_only,
        )


class RunExecuteStrategy(Protocol):
    def execute(self) -> RunResponse: ...


class RunExecuteMultipleStrategy(Protocol):
    def execute_all(self) -> List[RunResponse]: ...



"""Get HTTP headers needed to authenticate with the Platform API."""

import os
from typing import NamedTuple, Optional

import requests

from epx.run.exec.cloud.config import PLATFORM_API_VERSION
from epx import __version__ as EPX_CLIENT_VERSION


def platform_api_headers() -> dict[str, str]:
    """Headers needed to interact with the Platform API.

    Returns
    -------
    dict[str, str]
        Key/ value pairs to be passed as headers in requests to the Platform
        API.
    """
    auth_header = _get_auth_header()
    return {
        auth_header.key: f"Bearer {auth_header.value}",
        "content-type": "application/json",
        "fredcli-version": PLATFORM_API_VERSION,
        "user-agent": f"epx_client_{EPX_CLIENT_VERSION}",
    }


class _AuthHeader(NamedTuple):
    key: str
    value: str


def _get_auth_header() -> _AuthHeader:
    """Construct the authentication token portion of the header.

    There are two possible approaches to obtaining the auth token:
        1. Using a token refresher service
        2. Using an offline token set as the value of the
           `FRED_CLOUD_RUNNER_TOKEN` environment variable.

    This function attempts to obtain an auth token using option 1, and
    falls back to option 2 if a token cannot be obtained from the refresher
    service.

    Returns
    -------
    _AuthHeader
        Authentication token portion of the Platform API header.

    Raises
    ------
    RuntimeError
        If an authentication token cannot be obtained from either the
        token refresher service, or the local `FRED_CLOUD_RUNNER_TOKEN`
        environment variable.
    """
    if (refresher_header := _refresher_service_auth_header()) is not None:
        return refresher_header
    elif (offline_header := _offline_auth_header()) is not None:
        return offline_header
    else:
        raise RuntimeError("Cannot determine authorization token for Platform API.")


def _refresher_service_auth_header() -> Optional[_AuthHeader]:
    try:
        jupyter_token = os.environ["JPY_API_TOKEN"]
        endpoint = f'{os.environ["EPX_HUB_URL"]}services/refresher/tokens/access'
    except KeyError:
        return None
    response = requests.get(
        url=endpoint,
        headers={
            "Authorization": f"token {jupyter_token}",
        },
    )

    return _AuthHeader("Authorization", response.json().get("access_token"))


def _offline_auth_header() -> Optional[_AuthHeader]:
    """Return offline token using FRED_CLOUD_RUNNER_TOKEN env var if
    available. Otherwise return None.
    """
    try:
        return _AuthHeader("Offline-Token", os.environ["FRED_CLOUD_RUNNER_TOKEN"])
    except KeyError:
        return None



from abc import ABC, abstractmethod
import functools
from pathlib import Path
import re
from typing import Any, Callable, Iterable, Optional, Union, TypeGuard

import pandas as pd
import networkx as nx

from epx.run.fs import (
    CountType,
    FileFinderF11,
    FileFinderF10,
    FileFinder,
    VarBySimDayPath,
)


class RunResults(ABC):
    def __init__(self, file_finder: FileFinder):
        self._base_file_finder = file_finder

    def state(self, condition: str, state: str, count_type: CountType) -> pd.DataFrame:
        """Return a time series of agent occupancy of a given FRED state.

        Parameters
        ----------
        condition : str
            Name of the FRED model condition containing the target state.
        state : str
            Name of the target state.
        count_type : CountType
            Type of count to report. Options are:
                * `count`, the number of agents occupying the state at the end
                  of each simulated day.
                * `new`, the number of agents entering the state at any time
                  during each simulated day.
                * `cumulative`, the cumulative number of times any agent has
                  entered the state since the beginning of the simulation,
                  reported at the end of each simulated day.

        Returns
        -------
        pd.DataFrame
            DataFrame with columns:
                * `sim_day`
                * The name of the requested count type, i.e. one of `count`,
                  `new`, or `cumulative`.
        """
        return pd.read_csv(
            self._base_file_finder.state(condition, state, count_type),
            header=None,
            sep=" ",
            names=["sim_day", count_type],
        )

    def pop_size(self) -> pd.DataFrame:
        """Return a time series of population size during the run.

        Returns
        -------
        pd.DataFrame
            DataFrame with columns:
                * `sim_day`
                * `pop_size`.
        """
        return pd.read_csv(
            self._base_file_finder.pop_size,
            header=None,
            sep=" ",
            names=["sim_day", "pop_size"],
        )

    def epi_weeks(self) -> pd.DataFrame:
        """Return a mapping from simulation days to epidemiological weeks.

        An epidemiological week, commonly referred to as an epi week or a CDC
        week, is a standardized method of counting weeks to allow for the
        comparison of data year after year.

        By definition, the first epi week of the year ends on the first Saturday
        of January that falls at least four days into the month. Each epi week
        begins on a Sunday and ends on a Saturday.

        Returns
        -------
        pd.DataFrame
            DataFrame with columns:
                * `sim_day`
                * `epi_week` with string values in `YYYY.MM` format.
        """
        return pd.read_csv(
            self._base_file_finder.epi_week,
            header=None,
            sep=" ",
            names=["sim_day", "epi_week"],
            dtype={"sim_day": int, "epi_week": str},
        )

    def dates(self) -> pd.DataFrame:
        """Return a mapping from simulation days to calendar dates.

        Returns
        -------
        pd.DataFrame
            DataFrame with columns:
                * `sim_day`
                * `sim_date`
        """
        return pd.read_csv(
            self._base_file_finder.dates,
            header=None,
            sep=" ",
            names=["sim_day", "sim_date"],
            parse_dates=["sim_date"],
        )

    def print_output(self) -> pd.Series:
        """Return a series of messages output by FRED's `print()` action.

        Returns
        -------
        pd.Series
            A series of `print()` output messages.
        """
        with open(self._base_file_finder.print_output, "r") as f:
            return pd.Series([x.strip() for x in f.readlines()]).rename("print_output")

    def csv_output(self, filename: str) -> pd.DataFrame:
        """Return data output by FRED's `print_csv` action.

        Parameters
        ----------
        filename : str
            Name of output file as specified in the calls to `print_csv` in
            the FRED model code.

        Returns
        -------
        pd.DataFrame
            DataFrame containing data written to `filename`. Columns are those
            specified in the call to `open_csv` in the model code.
        """
        return pd.read_csv(self._base_file_finder.csv_output(filename))

    def file_output(self, filename: str) -> pd.Series:
        """Return data output by FRED's `print_file` action.

        Parameters
        ----------
        filename : str
            Name of output file as specified in the calls to `print_file` in
            the FRED model code.

        Returns
        -------
        pd.Series
            Series containing data written to `filename`. Each entry is a string
            containing the complete output for each call to `print_file` within
            the model code.
        """
        with open(self._base_file_finder.text_output(filename), "r") as f:
            return pd.Series([x.strip() for x in f.readlines()]).rename("file_output")

    def numeric_var(self, varname: str) -> pd.DataFrame:
        """Return a time series of the value of a shared numeric variable.

        Parameters
        ----------
        varname : str
            Name of the shared numeric variable.

        Returns
        -------
        pd.DataFrame
            DataFrame with columns:
                * `sim_day`
                * `value`
        """
        return pd.read_csv(
            self._base_file_finder.numeric(varname),
            header=None,
            sep=" ",
            names=["sim_day", "value"],
        )

    @abstractmethod
    def list_var(self, varname: str, wide: bool) -> pd.DataFrame:
        """Return a time series of the value of a shared list variable.

        Parameters
        ----------
        varname : str
            Name of the shared list variable.
        wide : bool, optional
            Return data in 'wide' format where a column is created for each
            list element. Defaults to ``False``.

        Returns
        -------
        pd.DataFrame
            DataFrame with columns:
                * `sim_day`
                * `list_index`
                * `value`
            if ``wide=False``. Otherwise:
                * `sim_day`
                * `item_0`
                * `item_1`
                * ...
        """
        ...

    @abstractmethod
    def list_table_var(self, varname: str, wide: bool) -> pd.DataFrame:
        """Return a time series of the value of a shared list_table variable.

        Parameters
        ----------
        varname : str
            Name of the shared list_table variable.
        wide : bool, optional
            Return data in 'wide' format where a column is created for each
            list element. Defaults to ``False``.

        Returns
        -------
        pd.DataFrame
            DataFrame with columns:
                * `sim_day`
                * `key`
                * `list_index`
                * `value`
            if ``wide=False``. Otherwise:
                * `sim_day`
                * `key`
                * `item_0`
                * `item_1`
                * ...
        """
        ...

    @abstractmethod
    def table_var(self, varname: str) -> pd.DataFrame:
        """Return a time series of the value of a shared table variable.

        Parameters
        ----------
        varname : str
            Name of the shared table variable.

        Returns
        -------
        pd.DataFrame
                * `sim_day`
                * `key`
                * `value`
        """
        ...

    @abstractmethod
    def network(
        self, network_name: str, is_directed: bool = True, sim_day: Optional[int] = None
    ) -> nx.Graph:
        """Reads a single day's network from the output files. Returns a networkx graph.

        Parameters
        ----------
        network_name : str
            Name of the network.
        is_directed : bool, optional
            Whether the network is directed or not. Defaults to ``True``.
        sim_day : int, optional
            The simulation day to read the network for. If not provided, the
            network for the last day of the simulation is read.

        Returns
        -------
        nx.Graph
        """
        ...

    @staticmethod
    def _str_to_numeric(s: str) -> Union[int, float]:
        def is_int(s):
            try:
                int(s)
                return True
            except ValueError:
                return False

        if is_int(s):
            return int(s)
        return float(s)

    @staticmethod
    def _read_table_file(filepath: Path) -> pd.DataFrame:
        """Read files with two columns where col1 is key and col2 is value."""
        return pd.read_csv(filepath, names=["key", "value"], skiprows=[0])

    @classmethod
    def _read_wide_list_file(cls, filepath: Path, value_pattern: str) -> pd.DataFrame:
        """Reads a file in a wide list format.

        These are files with structure:

            <variable>,value
            0,<list_string_repr>
            1,<list_string_repr>

        Where:
            * `<variable>` specifies what each row represents (e.g. `day` or
              `key`)
            * `<list_string_repr>` is a string representation of a list, e.g.
              `"[0.000000,0.000000,0.000000]"`, `13.200000,26.400000`, or
              `13.200000` where the last example is interpreted as a list of
              length 1.

        Parameters
        ----------
        filepath : Path
            Path to the file to read
        value_pattern : str
            A regular expression with a single capturing group that extracts a
            string of comma separated numbers from the <list_string_repr> (as
            defined above) appropriate for the given file. The idea is that
            hitting the extracted capturing group with `str.split(",")` will
            yield a list of strings representing numbers.

        Returns
        -------
        pd.DataFrame
            DataFrame with columns: `<variable>`, `list_index`, `value`.
        """

        def parse_line(line: str, value_pattern: str) -> tuple[Union[int, float], str]:
            m = re.match(r"([\d\.]*)," + value_pattern, line)
            if m is None:
                raise ValueError(f"Unrecognized wide list line: {line}")
            return cls._str_to_numeric(m.group(1)), m.group(2)

        def add_list_index_col(df):
            return (
                df.assign(tmp=1)
                .assign(list_index=lambda df: df["tmp"].cumsum() - 1)
                .drop(columns="tmp")
            )

        with open(filepath, "r") as f:
            col_names = [x.strip().lower() for x in f.readline().split(",")]
            col_names[1] = "value"
            lines = f.readlines()
            if not lines:
                data = None
            else:
                data = (parse_line(line, value_pattern) for line in lines)

        if data is None:
            return pd.DataFrame(columns=[col_names[0], "list_index", col_names[1]])
        else:
            return (
                pd.DataFrame.from_records(data, columns=col_names)
                .pipe(
                    lambda df: (
                        df.loc[:, [col_names[0]]].join(
                            df.loc[:, col_names[1]]
                            .str.split(",")
                            .explode()
                            .rename(col_names[1])
                            .astype(float)
                        )
                    )
                )
                .pipe(
                    lambda df: df.groupby(col_names[0])[col_names].apply(
                        add_list_index_col
                    )
                )
                .loc[:, [col_names[0], "list_index", col_names[1]]]
                .reset_index(drop=True)
            )

    @staticmethod
    def _add_wide_option(func):
        """Decorate a method to report list variables in 'wide' format.

        Long format data for a list variable, for example, looks like:

        .. code-block:: text

                sim_day  list_index  value
            0         0           0    0.0
            1         0           1    0.0
            2         0           2    0.0
            3         1           0    1.1
            4         1           1    2.2
            5         1           2    3.3

        This decorator looks for ``wraps=True`` among the keyword arguments
        passed to the decorated method and, if found, pivots the data into
        the form:

        .. code-block:: text

                sim_day  item_0  item_1  item_2
            0         0     0.0     0.0     0.0
            1         1     1.1     2.2     3.3

        Parameters
        ----------
        func : Callable[[RunResults, str], pd.DataFrame]
            Function that reads either a list or list_table variable in long
            format.

        Returns
        -------
        Callable[[RunResults, str], pd.DataFrame]
            Decorated method that implements pivot from long to wide data shape.
        """

        @functools.wraps(func)
        def inner(*args, **kwargs):
            if "wide" in kwargs.keys():
                wide = kwargs.pop("wide")
                if wide:
                    df = func(*args, **kwargs)
                    ix_cols = [x for x in df.columns if x != "value"]
                    return (
                        df.set_index(ix_cols)["value"]
                        .unstack()
                        .rename_axis(columns=None)
                        .rename(columns=lambda x: f"item_{x}")
                        .reset_index()
                    )
            return func(*args, **kwargs)

        return inner


class RunResultsF10(RunResults):
    def __init__(self, file_finder: FileFinderF10):
        super().__init__(file_finder)
        self._file_finder = file_finder

    @RunResults._add_wide_option
    def list_var(self, varname: str, wide: bool = False) -> pd.DataFrame:
        return self._read_complex_variable_time_series(
            varname,
            ["sim_day", "list_index", "value"],
            self._read_list_file,
            self._file_finder.list_by_simday,
            self._file_finder.list_end_of_sim,
        )

    @RunResults._add_wide_option
    def list_table_var(self, varname: str, wide: bool = False) -> pd.DataFrame:
        return self._read_complex_variable_time_series(
            varname,
            ["sim_day", "key", "list_index", "value"],
            functools.partial(self._read_wide_list_file, value_pattern=r"([\d,\.]*)"),
            self._file_finder.list_table_by_simday,
            self._file_finder.list_table_end_of_sim,
        )

    def table_var(self, varname: str) -> pd.DataFrame:
        return self._read_complex_variable_time_series(
            varname,
            ["sim_day", "key", "value"],
            self._read_table_file,
            self._file_finder.table_by_simday,
            self._file_finder.table_end_of_sim,
        )

    def network(
        self, network_name: str, is_directed: bool = True, sim_day: Optional[int] = None
    ) -> nx.Graph:
        if sim_day is None:
            sim_day = self.dates().iloc[-1]["sim_day"] + 1
        filepath = self._file_finder.network(network_name, sim_day)

        # create a Graph or DiGraph object
        G: Union[nx.DiGraph, nx.Graph]
        if is_directed:
            G = nx.DiGraph(name=network_name)
        else:
            G = nx.Graph(name=network_name)

        # read in data from file
        with open(filepath, "r") as f:
            lines = f.readlines()

        tie_data_index = lines.index("*tie data\n")
        node_data = lines[:tie_data_index]
        tie_data = lines[tie_data_index:]

        # construct a list of nodes
        node_count = 0
        nodes = []
        node_attr_keys = node_data[1].strip().split(" ")[1:]
        for line in node_data[2:]:
            line_list = line.strip().split(" ")
            node_id = line_list[0]
            node_attrs = line_list[1:]
            nodes.append((node_id, dict(zip(node_attr_keys, node_attrs))))
            node_count += 1

        # construct a list of ties
        tie_count = 0
        ties = []
        tie_attr_keys = tie_data[1].strip().split(" ")[2:]
        for line in tie_data[2:]:
            line_list = line.strip().split(" ")
            from_id = line_list[0]
            to_id = line_list[1]
            tie_attrs = line_list[2:]
            ties.append((from_id, to_id, dict(zip(tie_attr_keys, tie_attrs))))
            tie_count += 1

        G.add_nodes_from(nodes)
        G.add_edges_from(ties)

        return G

    def _read_complex_variable_time_series(
        self,
        varname: str,
        col_names: list[str],
        read_file: Callable[[Path], pd.DataFrame],
        interval_output_files: Callable[[str], Iterable[VarBySimDayPath]],
        end_of_sim_output_file: Callable[[str], Path],
    ) -> pd.DataFrame:
        """Read list, table, or list table as a time series.

        FRED 10 outputs two types of file for variables with complex data types:

        * {varname}-{sim_day}.txt if the variable is output with a specified
          interval
        * {varname}.txt at the end of the simulation (if this is turned on)

        This method uses the `read_file` callable to read all available
        individual files into data frames with `col_names`, and returns the data
        as a single data frame.

        Here we also handle the logic for:

        1. Using only end of simulation outputs, because interval outputs
           weren't turned on.
        2. Using only daily interval outputs, because end of simulation outputs
           weren't turned on
        3. Combining daily interval and end of simulation outputs in case both
           were turned on, and the last daily interval output occurred before
           the end of the simulation.
        """

        def outputs_by_simday(
            varname: str,
            col_names: list[str],
            read_file: Callable[[Path], pd.DataFrame],
            interval_output_files: Callable[[str], Iterable[VarBySimDayPath]],
        ) -> Optional[pd.DataFrame]:
            try:
                return (
                    pd.concat(
                        [
                            read_file(x.path).assign(sim_day=x.sim_day)
                            for x in interval_output_files(varname)
                        ]
                    )
                    .reset_index(drop=True)
                    .loc[:, col_names]
                )
            except FileNotFoundError:
                return None

        def outputs_end_of_sim(
            varname: str,
            col_names: list[str],
            last_simday: pd.Timedelta,
            read_file: Callable[[Path], pd.DataFrame],
            end_of_sim_output_file: Callable[[str], Path],
        ) -> Optional[pd.DataFrame]:
            try:
                return (
                    read_file(end_of_sim_output_file(varname))
                    .assign(sim_day=last_simday)
                    .reset_index(drop=True)
                    .loc[:, col_names]
                )
            except FileNotFoundError:
                return None

        df = outputs_by_simday(varname, col_names, read_file, interval_output_files)
        last_simday = self.dates().iloc[-1]["sim_day"]

        if df is None:
            df_end_of_sim = outputs_end_of_sim(
                varname, col_names, last_simday, read_file, end_of_sim_output_file
            )
            if df_end_of_sim is None:
                raise FileNotFoundError(
                    f"Could not find results data for variable: {varname}"
                )
            else:
                return df_end_of_sim
        else:
            if df["sim_day"].max() < last_simday:
                df_end_of_sim = outputs_end_of_sim(
                    varname, col_names, last_simday, read_file, end_of_sim_output_file
                )
                if df_end_of_sim is None:
                    return df
                else:
                    return pd.concat([df, df_end_of_sim]).reset_index(drop=True)
            else:
                return df

    @staticmethod
    def _read_list_file(filepath: Path) -> pd.DataFrame:
        return (
            pd.read_csv(filepath, names=["value"], skiprows=[0])
            # Create list_index column
            .assign(tmp=1)
            .assign(list_index=lambda df: df["tmp"].cumsum() - 1)
            .drop(columns="tmp")
        )

    def __repr__(self):
        return f"RunResultsF10({self._file_finder})"


class RunResultsF11(RunResults):
    def __init__(self, file_finder: FileFinderF11):
        super().__init__(file_finder)
        self._file_finder = file_finder

    @RunResults._add_wide_option
    def list_var(self, varname: str, wide: bool = False) -> pd.DataFrame:
        return self._read_wide_list_file(
            self._file_finder.list_(varname), r'"\[([\d,\.]*)\]"'
        ).rename(columns={"day": "sim_day"})

    @RunResults._add_wide_option
    def list_table_var(self, varname: str, wide: bool = False) -> pd.DataFrame:
        return (
            pd.concat(
                [
                    (
                        self._read_wide_list_file(v.path, r'"\[([\d,\.]*)\]"')
                        .rename(columns={"day": "sim_day"})
                        .assign(sim_day=v.sim_day)
                    )
                    for v in self._file_finder.list_table_by_simday(varname)
                ]
            )
            .reset_index(drop=True)
            .loc[:, ["sim_day", "key", "list_index", "value"]]
        )

    def table_var(self, varname: str):
        return (
            pd.concat(
                [
                    self._read_table_file(v.path).assign(sim_day=v.sim_day)
                    for v in self._file_finder.table_by_simday(varname)
                ]
            )
            .reset_index(drop=True)
            .loc[:, ["sim_day", "key", "value"]]
        )

    def network(
        self, network_name: str, is_directed: bool = True, sim_day: Optional[int] = None
    ) -> nx.Graph:
        if sim_day is None:
            sim_day = self.dates().iloc[-1]["sim_day"]
        filepath = self._file_finder.network(network_name, sim_day)
        file_data = nx.nx_pydot.read_dot(filepath)
        # remove quotes
        for node in file_data.nodes:
            if "id" in file_data.nodes[node]:
                file_data.nodes[node]["id"] = file_data.nodes[node]["id"].strip('"')
        # When support for FRED 10 is dropped, we will no longer need
        # the is_directed parameter and can use the following code instead:
        # if isinstance(file_data, nx.MultiDiGraph):
        if is_directed:
            return nx.DiGraph(file_data, name=network_name)
        else:
            return nx.Graph(file_data, name=network_name)

    def __repr__(self):
        return f"RunResultsF11({self._file_finder})"


class RunResultsFactory:
    def __init__(self, file_finder: Union[FileFinderF10, FileFinderF11]):
        self.file_finder = file_finder

    def build(self) -> RunResults:
        """Obtain a RunStatus appropriate for the given output directory.

        Raises
        ------
        TypeError
            If given `file_finder` is not a recognized type.
        """
        # TODO: Reduce code duplication across this and RunStatusFactory
        if self._is_legacy_file_finder(self.file_finder):
            return RunResultsF10(self.file_finder)
        elif self._is_file_finder(self.file_finder):
            return RunResultsF11(self.file_finder)
        else:
            raise TypeError(f"{self.file_finder} is not a valid FileFinder type")

    @staticmethod
    def _is_legacy_file_finder(file_finder: Any) -> TypeGuard[FileFinderF10]:
        if isinstance(file_finder, FileFinderF10):
            return True
        return False

    @staticmethod
    def _is_file_finder(file_finder: Any) -> TypeGuard[FileFinderF11]:
        if isinstance(file_finder, FileFinderF11):
            return True
        return False
